#!/usr/bin/env python3
# ....................{ LICENSE                           }....................
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS                           }....................
import numpy as np
from betse.util.app.meta import appmetaone
from betse.exceptions import BetseSimConfException
from betse.lib.matplotlib import mplcolormap
from betse.lib.yaml import yamls
from betse.lib.yaml.yamlalias import yaml_alias, yaml_enum_alias
from betse.lib.yaml.abc.yamlfileabc import YamlFileDefaultABC
from betse.science.enum.enumconf import (
    CellLatticeType, GrnUnpicklePhaseType, IonProfileType, SolverType)
from betse.science.config.model.conftis import (
    SimConfCutListItem, SimConfTissueDefault, SimConfTissueListItem)
# from betse.util.io.log import logs
from betse.util.path import dirs, pathnames
from betse.util.type.descriptor.descs import classproperty_readonly
from betse.util.type.types import IterableTypes, SequenceTypes, StrOrNoneTypes

# ....................{ SUBCLASSES                        }....................
class Parameters(YamlFileDefaultABC):
    '''
    Root YAML-backed in-memory and on-disk simulation configuration,
    encapsulating a low-level container of simulation configuration settings
    both loaded from and saved back to a YAML-formatted configuration file.

    Attributes (Solver)
    ----------
    solver_type : SolverType
        Type of **simulation solver** (i.e., numerical technique iteratively
        computing simulation time steps) with which to solve this simulation.

    Attributes (Exports)
    ----------
    anim : SimConfExportAnims
        Subconfiguration configuring exported animations.
    csv : SimConfExportCSVs
        Subconfiguration configuring exported comma-separated value (CSV)
        files.
    plot : SimConfExportPlots
        Subconfiguration configuring exported plots.
    visual : SimConfExportVisual
        Subconfiguration applicable to *all* exported visuals (i.e.,
        animations, plots).

    Attributes (Exports: Colormaps)
    ----------
    colormap_diverging_name : str
        Matplotlib-specific name of the default **diverging colormap** (i.e.,
        object mapping zero-centered cell data deviating both above and below 0
        in equal measure to color values in visual exports). Since most cell
        data is zero-centered, this is effectively the default colormap.
    colormap_sequential_name : str
        Matplotlib-specific name of the default **sequential colormap** (i.e.,
        object mapping zero-based cell data monotonically increasing from 0 to
        positive infinity to color values in visual exports). Since most
        vectorial cell data (e.g., deformations, fluid flow, electromagnetic
        currents and fields) is zero-based, this is effectively the default
        colormap for such data.
    colormap_gj_name : str
        Matplotlib-specific name of the default **gap junction colormap**
        (i.e., object mapping gap junction data, including currents, to color
        values in visual exports).
    colormap_grn_name : str
        Matplotlib-specific name of the default **gene regulatory network (GRN)
        colormap** (i.e., object mapping GRN data, including concentration
        nodes of reaction network graphs, to color values in visual exports).

    Attributes (Path: Export)
    ----------
    init_export_dirname : str
        Absolute pathname of the directory containing results exported from
        this simulation's most recent initialization run, guaranteed to exist.
    init_export_dirname_relative : str
        Relative pathname of the directory containing results exported from
        this simulation's most recent initialization run, relative to the
        absolute path of the directory containing this simulation
        configuration's YAML file.
    sim_export_dirname : str
        Absolute pathname of the directory containing results exported from
        this simulation's most recent simulation run, guaranteed to exist.
    sim_export_dirname_relative : str
        Relative pathname of the directory containing results exported from
        this simulation's most recent simulation run, relative to the absolute
        path of the directory containing this simulation configuration's YAML
        file.

    Attributes (Path: Pickle: Seed)
    ----------
    seed_pickle_basename : str
        Basename of the pickled file providing this simulation's most recently
        seeded cell cluster, relative to the :attr:`init_pickle_dirname`
        directory.
    seed_pickle_filename : str
        Abolute filename of the pickled file providing this simulation's most
        recently seeded cell cluster.

    Attributes (Path: Pickle: Initialization)
    ----------
    init_pickle_basename : str
        Basename of the pickled file providing this simulation's most recent
        initialization run, relative to the :attr:`init_pickle_dirname`
        directory.
    init_pickle_filename : str
        Abolute filename of the pickled file providing this simulation's most
        recent initialization run.
    init_pickle_dirname : str
        Absolute dirname of the directory containing this simulation's most
        recently seeded cell cluster *and* initialization run, guaranteed to
        exist.
    init_pickle_dirname_relative : str
        Relative dirname of the directory containing this simulation's most
        recently seeded cell cluster *and* initialization run, relative to the
        :attr:`conf_dirname` directory.

    Attributes (Path: Pickle: Simulation)
    ----------
    sim_pickle_basename : str
        Basename of the pickled file providing this simulation's most recent
        simulation run within the current :attr:`sim_dirname`.
    sim_pickle_filename : str
        Abolute filename of the pickled file providing this simulation's most
        recent simulation run.
    sim_pickle_dirname : str
        Absolute dirname of the directory containing this simulation's most
        recent simulation run, guaranteed to exist.
    sim_pickle_dirname_relative : str
        Relative dirname of the directory containing this simulation's most
        recent simulation run, relative to the :attr:`conf_dirname` directory.

    Attributes (Path: Pickle: Gene)
    ----------
    grn_pickle_basename : str
        Basename of the pickled file providing this simulation's most recent
        gene regulatory network (GRN) run, relative to the
        :attr:`grn_pickle_dirname` directory.
    grn_pickle_filename : str
        Absolute filename of the pickled file providing this simulation's most
        recent gene regulatory network (GRN) run.
    grn_pickle_dirname : str
        Absolute dirname of the directory containing this simulation's most
        recent gene regulatory network (GRN) run, guaranteed to exist.
    grn_pickle_dirname_relative : str
        Relative dirname of the directory containing this simulation's most
        recent gene regulatory network (GRN) run, relative to the
        :attr:`conf_dirname` directory.
    grn_unpickle_filename : StrOrNoneTypes
        Absolute filename of the pickled file providing a prior gene regulatory
        network (GRN) run for this simulation if restarting the current such
        run from where this prior run left off *or* ``None`` otherwise (i.e.,
        if starting the current such run from scratch).
    grn_unpickle_filename_relative : StrOrNoneTypes
        Relative filename of the pickled file providing a prior gene regulatory
        network (GRN) run for this simulation, relative to the
        :attr:`conf_dirname` directory, if restarting the current such run from
        where this prior run left off *or* ``None`` otherwise (i.e., if
        starting the current such run from scratch).

    Attributes (Space: Cell)
    ----------
    cell_radius : float
        Radius in meters of each cell in this cluster. This should typically be
        within an order of magnitude of the recommended default of ``5.0e-6``.

    Attributes (Space: Cell Cluster)
    ----------
    cell_lattice_disorder : float
        Degree to which cell boundaries spatially deviate from the base cell
        lattice. If 0.0, cells rigidly conform to this lattice; else, cells
        randomly deviate from this lattice. Increasing this increases the
        entropy (i.e., randomness) of these deviations, producing an
        increasingly chaotic arrangement of cells. This should typically reside
        in the range ``[0.0, 0.8]``.
    cell_lattice_type : CellLatticeType
        Type of **base cell lattice** (i.e., uniform grid to which cells are
        situated *before* random lattice disorder is applied).

    Attributes (Space: Environment)
    ----------
    grid_size : int
        Number of square grid spaces (in both the X and Y dimensions) to
        computationally divide this simulation's square environment into.
        Increasing this increases simulation granularity and hence stability at
        a quadratic increase in space and time costs (i.e., ``O(grid_size **
        2)``). This should typically reside in the range ``[10, 60]``.
    is_ecm : bool
        ``True`` only if the extracellular matrix (ECM) simulating
        **extracellular spaces** (i.e., the environment surrounding each cell
        in the cluster) is enabled. Disabling this reduces simulation accuracy
        at a substantial reduction in space and time costs.
    world_len : float
        Length in meters of both the X and Y dimensions of this simulation's
        square environment. This should typically reside in the range ``[80e-6,
        1000e-6]``.

    Attributes (Space: Tissue)
    ----------
    is_tissue_profiles : bool
        ``True`` only if **tissue profiles** (i.e., user-defined regions within
        the cell cluster to which specific base membrane diffusion profiles,
        interventions, and individualized dynamics may be applied) are enabled.
        Note that tissue profiles should typically *always* be enabled.
    tissue_default : SimConfTissueDefault
        Default tissue profile applied to all cells *not* already targeted by
        another tissue profile in the :attr:`tissue_profiles` list.
    tissue_profiles : YamlList
        List of all **non-default tissue profiles** (i.e., objects targeting a
        region of the cell cluster to be associated with particular simulation
        constants and parameters). Ignored if :attr:`is_tissue_profiles` is
        ``False``.

    Attributes (Space: Tissue: Cut)
    ----------
    cut_profiles : YamlList
        List of all **cut profiles** (i.e., objects targeting a region of the
        cell cluster to be permanently removed by a corresponding simulation
        event). Ignored if :attr:`is_tissue_profiles` is ``False``.
    event_cut_profile_names : SequenceTypes
        List of the names of all cut profiles whose corresponding cells are to
        be removed by the cutting event. Cut profiles whose names are *not* in
        this list are effectively ignored.
    event_cut_time : float
        Time step of the simulation phase at which to apply the cutting event.

    Attributes (Time: Total)
    ----------
    init_time_total : float
        Duration in seconds of the initialization phase.
    sim_time_total : float
        Duration in seconds of the simulation phase.

    Attributes (Time: Step)
    ----------
    init_time_step : float
        Duration in seconds of each time step (including both sampled and
        unsampled) for the initialization phase.
    init_tsteps : int
        Total number of all time steps (including both sampled and unsampled)
        for the initialization phase.
    sim_time_step : float
        Duration in seconds of each time step (including both sampled and
        unsampled) for the simulation phase.
    sim_tsteps : int
        Total number of all time steps (including both sampled and unsampled)
        for the simulation phase.

    Attributes (Time: Sampling)
    ----------
    init_time_sampling : float
        Duration in seconds between each sampled time step (including that
        sampled time step itself) for the initialization phase. Decreasing this
        duration increases the number of time steps for which data is exported
        from this phase at a linear cost in space consumption.
    sim_time_sampling : float
        Duration in seconds between each sampled time step (including that
        sampled time step itself) for the simulation phase. Decreasing this
        duration increases the number of time steps for which data is exported
        from this phase at a linear cost in space consumption.

    Attributes (Gene Regulatory Network)
    ----------
    The following attributes are ignored if :attr:`grn_enabled` is ``False``.

    grn : SimConfGrnFile
        Gene regulatory network (GRN) subconfiguration, encapsulating *all*
        GRN-related settings both loaded from and saved back to the separate
        YAML-formatted configuration file with filename
        :attr:`grn_config_filename`.
    grn_unpickle_phase_type : GrnUnpicklePhaseType
        Type of **unpickle simulation phase** (i.e., previously pickled
        simulation phase to unpickle as the computational basis for the current
        network to be run by the ``betse sim-grn`` subcommand).

    Attributes (Ion)
    ----------
    ion_profile : IonProfileType
        Type of **ion profile** (i.e., predefined set of all extracellular and
        cytosolic ions enabled by this simulation).
    ions_dict : dict
        Dictionary mapping:

        * From each of the following names of a supported core ion:

          * ``Na``, the sodium cation Na+.
          * ``K``, the potassium cation K+.
          * ``Ca``, the calcium cation Ca2+.
          * ``Cl``, the chloride anion Cl-.
          * ``M``, the unidentified anion M-.
          * ``P``, the anionic protein P-.

        * To either:

          * 0, if that ion is disabled by the ion profile in the current
            simulation configuration.
          * 1, if that ion is enabled by that ion profile.

    Attributes (Ion: Initial)
    ----------
    conc_env_ca : float
        Initial extracellular concentration of the Ca2+ (calcium) cation.
    cCl_env : float
        Initial extracellular concentration of the Cl- (chloride) anion.
    conc_env_k : float
        Initial extracellular concentration of the K+ (potassium) cation.
    conc_env_m : float
        Initial extracellular concentration of the M- (unidentified) anion,
        synthetically manufactured to enforce charge-balancing over all
        environmental spaces for both simulation stability and correctness.
    conc_env_na : float
        Initial extracellular concentration of the Na+ (sodium) cation.
    conc_env_p : float
        Initial extracellular concentration of the P- (protein) anion.

    Attributes (Ion: Initial: Custom)
    ----------

    Attributes (General: Scalars)
    ----------
    cell_polarizability : NumericSimpleTypes
        Constant defining the rate of cell polarizability change in electric
        fields, typically ranging ``[0.0, 1.0e-3]``.
    '''

    # ..................{ ALIASES                           }..................
    # To sanitize computation throughout the codebase, *ALL* real numbers are
    # required to be floating point rather the more general
    # "NumericSimpleTypes" type (i.e., either floating point or integer). Due
    # to magic internal to the yaml_alias() data descriptor, integer values are
    # both silently and safely cast to floating point values.

    # ..................{ ALIASES ~ solver                  }..................
    solver_type = yaml_enum_alias("['solver options']['type']", SolverType)

    # ..................{ ALIASES ~ export : colormap       }..................
    #FIXME: Define a new yaml_set_alias() data descriptor constraining the
    #value of the passed YAML key to a finite set of permissible values -- in
    #this case, colormap names in the set returned by the
    #betse.lib.matplotlib.mplcolormap.iter_colormap_names() iterator. For now,
    #we avoid constraining these names for simplicity.
    colormap_diverging_name = yaml_alias(
        "['results options']['default colormap']", str)
    colormap_sequential_name = yaml_alias(
        "['results options']['background colormap']", str)
    colormap_gj_name = yaml_alias(
        "['results options']['gj colormap']", str)
    colormap_grn_name = yaml_alias(
        "['results options']['network colormap']", str)

    # ..................{ ALIASES ~ path : seed             }..................
    seed_pickle_basename = yaml_alias("['init file saving']['worldfile']", str)

    # ..................{ ALIASES ~ path : init             }..................
    init_pickle_basename = yaml_alias("['init file saving']['file']", str)
    init_pickle_dirname_relative = yaml_alias(
        "['init file saving']['directory']", str)
    init_export_dirname_relative = yaml_alias(
        "['results file saving']['init directory']", str)

    # ..................{ ALIASES ~ path : sim              }..................
    sim_pickle_basename = yaml_alias("['sim file saving']['file']", str)
    sim_pickle_dirname_relative = yaml_alias(
        "['sim file saving']['directory']", str)
    sim_export_dirname_relative  = yaml_alias(
        "['results file saving']['sim directory']", str)

    # ..................{ ALIASES ~ path : grn              }..................
    grn_pickle_basename = yaml_alias(
        "['gene regulatory network settings']"
        "['sim-grn settings']['save to file']", str)
    grn_pickle_dirname_relative = yaml_alias(
        "['gene regulatory network settings']"
        "['sim-grn settings']['save to directory']", str)
    grn_unpickle_filename_relative = yaml_alias(
        "['gene regulatory network settings']"
        "['sim-grn settings']['load from']", StrOrNoneTypes)

    # ..................{ ALIASES ~ space : cell            }..................
    cell_radius = yaml_alias("['world options']['cell radius']", float)

    # ..................{ ALIASES ~ space : cell cluster    }..................
    cell_lattice_disorder = yaml_alias(
        "['world options']['lattice disorder']", float)
    cell_lattice_type = yaml_enum_alias(
        "['world options']['lattice type']", CellLatticeType)

    # ..................{ ALIASES ~ space : env             }..................
    grid_size = yaml_alias("['general options']['comp grid size']", int)
    is_ecm = yaml_alias(
        "['general options']['simulate extracellular spaces']", bool)
    world_len = yaml_alias("['world options']['world size']", float)

    # ..................{ ALIASES ~ space : tissue          }..................
    #FIXME: Does this boolean actually serve a demonstrable purpose? I might be
    #offbase here, but don't we always want tissue profiles? Is there actually
    #a useful use case for even disabling all tissue profiles?
    is_tissue_profiles = yaml_alias(
        "['tissue profile definition']['profiles enabled']", bool)

    # ..................{ ALIASES ~ time : init             }..................
    init_time_total = yaml_alias("['init time settings']['total time']", float)
    init_time_step = yaml_alias("['init time settings']['time step']", float)
    init_time_sampling = yaml_alias(
        "['init time settings']['sampling rate']", float)

    # ..................{ ALIASES ~ time : sim              }..................
    sim_time_total  = yaml_alias("['sim time settings']['total time']", float)
    sim_time_step  = yaml_alias("['sim time settings']['time step']", float)
    sim_time_sampling = yaml_alias(
        "['sim time settings']['sampling rate']", float)

    # ..................{ ALIASES ~ grn                     }..................
    grn_unpickle_phase_type = yaml_enum_alias(
        "['gene regulatory network settings']"
        "['sim-grn settings']['run network on']",
        GrnUnpicklePhaseType)

    # ..................{ ALIASES ~ ion                     }..................
    #FIXME: Consider shifting all ion-centric functionality into a dedicated
    #"ion" instance variable, instantiated to be an instance of a newly defined
    #"SimConfIonProfile" class. Ion handling currently appears to consume in
    #upwards of a third of this entire submodule.
    ion_profile = yaml_enum_alias(
        "['general options']['ion profile']", IonProfileType)

    # ..................{ ALIASES ~ ion ~ custom            }..................
    #FIXME: Actually use this.
    ion_profile_custom_conc_env_na = yaml_alias(
        "['general options']['customized ion profile']"
        "['extracellular Na+ concentration']", float)

    # ..................{ ALIASES ~ scalar                  }..................
    cell_polarizability = yaml_alias(
        "['internal parameters']['cell polarizability']", float)

    # ..................{ PROPERTIES                        }..................
    @classproperty_readonly
    def conf_default_filename(cls) -> str:
        return appmetaone.get_app_meta().betse_sim_conf_default_filename

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:

        # Method-specific imports aggregated here for maintainability.
        from betse.science.config.export.confexpcsv import SimConfExportCSVs
        from betse.science.config.export.visual.confexpvisanim import (
            SimConfExportAnims)
        from betse.science.config.export.visual.confexpvisplot import (
            SimConfExportPlots)
        from betse.science.config.export.visual.confexpvisual import (
            SimConfExportVisual)
        from betse.science.config.grn.confgrn import SimConfGrnFile

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Nullify all instance variables for safety.
        self._unload_paths()

        # Classify unloaded tissue and cut profiles.
        self.cut_profiles = SimConfCutListItem.make_list()
        self.tissue_default = SimConfTissueDefault()
        self.tissue_profiles = SimConfTissueListItem.make_list()

        # Classify unloaded export subconfigurations.
        self.anim = SimConfExportAnims()
        self.csv  = SimConfExportCSVs()
        self.plot = SimConfExportPlots()
        self.visual = SimConfExportVisual()

        # Classify unloaded GRN subconfigurations.
        self.grn = SimConfGrnFile()

    # ..................{ LOADERS                           }..................
    #FIXME: Convert all or most of the variables parsed by this method into
    #aliases of the above form. Brainy rainbows!
    def load(self, *args, **kwargs) -> None:

        # Avoid circular import dependencies.
        from betse.science.compat import compatconf

        # Version of the YAML specification this file is implicitly assumed to
        # comply with, preserving backward compatibility with older files
        # erroneously prefaced by the "%YAML 1.1" directive.
        YAML_VERSION = '1.2'

        # Load this file under the typically safe assumption this file complies
        # with the YAML 1.2 specification, preserving backward compatibility
        # with older files erroneously prefaced by the "%YAML 1.1" directive.
        super().load(*args, yaml_version=YAML_VERSION, **kwargs)

        # Preserve backward compatibility with prior configuration formats
        # *BEFORE* other initialization, which expects the passed YAML file to
        # conform to the current configuration format.
        compatconf.upgrade_sim_conf(self)

        # Initialize paths specified by this configuration.
        self.reload_paths()

        # Load all currently unloaded tissue subconfigurations.
        self.cut_profiles.load(
            conf=self._conf['tissue profile definition']['cut profiles'])
        self.tissue_default.load(
            conf=self._conf['tissue profile definition']['tissue']['default'])
        self.tissue_profiles.load(
            conf=self._conf['tissue profile definition']['tissue']['profiles'])

        # Load all currently unloaded export subconfigurations.
        self.anim  .load(conf=self._conf)
        self.csv   .load(conf=self._conf)
        self.plot  .load(conf=self._conf)
        self.visual.load(conf=self._conf)

        #----------------------------------------------------------------------
        # GENERAL OPTIONS
        #----------------------------------------------------------------------

        self.autoInit = self._conf['automatically run initialization']
        self.plot_grid_size = 50

        # Total number of time steps for the initialization and simulation
        # phases, rounding down to the nearest time step to preserve sanity.
        self.init_tsteps = int(self.init_time_total / self.init_time_step)
        self. sim_tsteps = int(self. sim_time_total / self. sim_time_step)

        #----------------------------------------------------------------------
        # WORLD OPTIONS
        #----------------------------------------------------------------------

        # Geometric constants and factors
        self.wsx = self.world_len  # the x-dimension of the world space [m]
        self.wsy = self.world_len  # the y-dimension of the world space [m]
        self.cell_height = float(self._conf['world options']['cell height'])  # the height of a cell in the z-direction
        self.cell_space = float(self._conf['world options']['cell spacing'])  # the true cell-cell spacing

        volmult = float(self._conf['internal parameters']['environment volume multiplier'])

        self.vol_env = volmult*self.wsx*self.wsy*self.cell_height  # environmental volume for "no ECM" simulation

        # Parameters for Lloyd's Voronoi mesh optimization settings during seed:
        # FIXME need to be put into betse.science.compat:
        mesh_refine = self._conf['world options'].get('mesh refinement', None)

        if mesh_refine is not None:
            self.refine_mesh = mesh_refine['refine mesh']
            self.maximum_voronoi_steps = int(mesh_refine['maximum steps'])
            self.voronoi_convergence = float(mesh_refine['convergence threshold'])

        else:
            self.refine_mesh = False
            self.maximum_voronoi_steps = 10
            self.voronoi_convergence = 2.5

        # Parameters for import of cell seed centers and clipping curve from user-defined svg files:
        # FIXME need to be put into betse.science.compat:
        svg_import = self._conf['world options'].get('import from svg', None)

        if svg_import is not None:

            self.svg_size = svg_import['svg size']
            self.svg_override = svg_import['svg override']
            svg_cells_fname_o = svg_import['cells from svg']

            self.svg_cells_fname = pathnames.join(self.conf_dirname, svg_cells_fname_o)


        else:
            self.svg_override = False
            self.svg_cells_fname = None
            self.svg_size = None

        # simulate only a single cell # FIXME need to be put into betse.science.compat:
        self.single_cell = self._conf['world options'].get('simulate single cell', False)

        # define the alpha shape threshholds # FIXME need to be put into betse.science.compat:
        self.alpha_shape = self._conf['world options'].get('alpha shape', 0.05)

        # allow user to use geometric centers of mesh triangles or circumcenters:
        # FIXME need to be put into betse.science.compat:
        self.use_centroids = self._conf['world options'].get('use centers', False)


        #---------------------------------------------------------------------------------------------------------------
        # TARGETED INTERVENTIONS
        #---------------------------------------------------------------------------------------------------------------

        #FIXME: This dictionary appears to be superfluous. It's never iterated
        #over or treated like a dictionary. Its keys are only ever looked up. In
        #short, this dictionary shouldn't exist. For efficiency and sanity, all
        #keys of this dictionary should be refactored into simple instance
        #variables of this class (e.g., from "p.scheduled_options['Na_mem']" to
        #simply "p.event_Na_mem").
        #FIXME: After doing so, these instance variables should be refactored
        #from unreadable lists to objects with human-readable attribute names
        #(e.g., from "p.event_Na_mem[0]" to "p.event_Na_mem.start_time").
        #
        #The longest journey to the future begins tomorrow.

        # initialize dictionary keeping track of targeted scheduled options for the sim:
        self.scheduled_options = {}

        bool_Namem = bool(self._conf['change Na mem']['event happens'])
        bool_Kmem = bool(self._conf['change K mem']['event happens'])
        bool_Clmem = bool(self._conf['change Cl mem']['event happens'])
        bool_Camem = bool(self._conf['change Ca mem']['event happens'])
        bool_ecmj = bool(self._conf['break ecm junctions']['event happens'])
        bool_pressure = bool(self._conf['apply pressure']['event happens'])

        if bool_Namem:
            on_Namem = float(self._conf['change Na mem']['change start'])
            off_Namem = float(self._conf['change Na mem']['change finish'])
            rate_Namem = float(self._conf['change Na mem']['change rate'])
            multi_Namem = float(self._conf['change Na mem']['multiplier'])
            apply_Namem = self._conf['change Na mem']['apply to']
            function = self._conf['change Na mem']['modulator function']
            Namem = [on_Namem, off_Namem, rate_Namem, multi_Namem, apply_Namem,function]
            self.scheduled_options['Na_mem'] = Namem
        else:
            self.scheduled_options['Na_mem'] = 0

        if bool_Kmem:
            on_Kmem = float(self._conf['change K mem']['change start'])
            off_Kmem = float(self._conf['change K mem']['change finish'])
            rate_Kmem = float(self._conf['change K mem']['change rate'])
            multi_Kmem = float(self._conf['change K mem']['multiplier'])
            apply_Kmem = self._conf['change K mem']['apply to']
            function = self._conf['change K mem']['modulator function']
            Kmem = [on_Kmem, off_Kmem, rate_Kmem, multi_Kmem, apply_Kmem,function]
            self.scheduled_options['K_mem'] = Kmem
        else:
            self.scheduled_options['K_mem'] = 0

        if bool_Clmem:
            on_Clmem = float(self._conf['change Cl mem']['change start'])
            off_Clmem = float(self._conf['change Cl mem']['change finish'])
            rate_Clmem = float(self._conf['change Cl mem']['change rate'])
            multi_Clmem = float(self._conf['change Cl mem']['multiplier'])
            apply_Clmem = self._conf['change Cl mem']['apply to']
            function = self._conf['change Cl mem']['modulator function']
            Clmem = [on_Clmem, off_Clmem, rate_Clmem, multi_Clmem, apply_Clmem, function]
            self.scheduled_options['Cl_mem'] = Clmem
        else:
            self.scheduled_options['Cl_mem'] = 0

        if bool_Camem:
            on_Camem = float(self._conf['change Ca mem']['change start'])
            off_Camem = float(self._conf['change Ca mem']['change finish'])
            rate_Camem = float(self._conf['change Ca mem']['change rate'])
            multi_Camem = float(self._conf['change Ca mem']['multiplier'])
            apply_Camem = self._conf['change Ca mem']['apply to']
            function = self._conf['change Ca mem']['modulator function']
            Camem = [on_Camem, off_Camem, rate_Camem, multi_Camem, apply_Camem,function]
            self.scheduled_options['Ca_mem'] = Camem
        else:
            self.scheduled_options['Ca_mem'] = 0

        if bool_pressure:
            on_p = float(self._conf['apply pressure']['change start'])
            off_p = float(self._conf['apply pressure']['change finish'])
            rate_p = float(self._conf['apply pressure']['change rate'])
            multi_p = float(self._conf['apply pressure']['multiplier'])
            apply_p = self._conf['apply pressure']['apply to']
            function = self._conf['apply pressure']['modulator function']
            pressure_ops = [on_p, off_p, rate_p, multi_p, apply_p,function]

            self.scheduled_options['pressure'] = pressure_ops
        else:
            self.scheduled_options['pressure'] = 0

        if bool_ecmj:
            on_ecmj = float(self._conf['break ecm junctions']['change start'])
            off_ecmj = float(self._conf['break ecm junctions']['change finish'])
            rate_ecmj = float(self._conf['break ecm junctions']['change rate'])
            apply_ecmj = self._conf['break ecm junctions']['apply to']
            mult_ecmj = 1.0 - float(self._conf['break ecm junctions'].get('multiplier', 0.0))
            self.scheduled_options['ecmJ'] = [
                on_ecmj, off_ecmj, rate_ecmj, apply_ecmj, mult_ecmj]
        else:
            self.scheduled_options['ecmJ'] = 0

        # Parameterize the cutting event if enabled.
        self.break_TJ = self._conf['cutting event'].get('break TJ', True)
        self.wound_TJ = float(self._conf['cutting event'].get('wound TJ', 0.1))
        # self.event_cut_time = float(self._conf['cutting event'].get('cut time', 0.0))
        self.event_cut_time = 0.0
        self.event_cut_profile_names = self._conf['cutting event']['apply to']

        #---------------------------------------------------------------------------------------------------------------
        # GLOBAL INTERVENTIONS
        #---------------------------------------------------------------------------------------------------------------

        # initialize dictionary keeping track of global scheduled options for the sim:
        self.global_options = {}

        bool_Naenv = bool(self._conf['change Na env']['event happens'])
        bool_Kenv = bool(self._conf['change K env']['event happens'])
        bool_Clenv = bool(self._conf['change Cl env']['event happens'])
        bool_gjblock = bool(self._conf['block gap junctions']['event happens'])
        bool_temp =  bool(self._conf['change temperature']['event happens'])
        bool_NaKblock = bool(self._conf['block NaKATP pump']['event happens'])

        if bool_Kenv is False:
            self.global_options['K_env'] = 0
        elif bool_Kenv is True:
            on_Kenv = float(self._conf['change K env']['change start'])
            off_Kenv = float(self._conf['change K env']['change finish'])
            rate_Kenv = float(self._conf['change K env']['change rate'])
            multi_Kenv = float(self._conf['change K env']['multiplier'])
            kenv = [on_Kenv, off_Kenv, rate_Kenv, multi_Kenv]
            self.global_options['K_env'] = kenv

        if bool_Clenv is False:
            self.global_options['Cl_env'] = 0
        elif bool_Clenv is True:
            on_Clenv = float(self._conf['change Cl env']['change start'])
            off_Clenv = float(self._conf['change Cl env']['change finish'])
            rate_Clenv = float(self._conf['change Cl env']['change rate'])
            multi_Clenv = float(self._conf['change Cl env']['multiplier'])
            Clenv = [on_Clenv, off_Clenv, rate_Clenv, multi_Clenv]
            self.global_options['Cl_env'] = Clenv

        if bool_Naenv is False:
            self.global_options['Na_env'] = 0
        elif bool_Naenv is True:
            on_Naenv = float(self._conf['change Na env']['change start'])
            off_Naenv = float(self._conf['change Na env']['change finish'])
            rate_Naenv = float(self._conf['change Na env']['change rate'])
            multi_Naenv = float(self._conf['change Na env']['multiplier'])
            Naenv = [on_Naenv, off_Naenv, rate_Naenv, multi_Naenv]
            self.global_options['Na_env'] = Naenv

        if bool_gjblock is False:
            self.global_options['gj_block'] = 0
        elif bool_gjblock is True:
            on_gj = float(self._conf['block gap junctions']['change start'])
            off_gj = float(self._conf['block gap junctions']['change finish'])
            rate_gj = float(self._conf['block gap junctions']['change rate'])
            fraction_gj = float(self._conf['block gap junctions']['random fraction'])
            gjb = [on_gj,off_gj,rate_gj,fraction_gj]
            self.global_options['gj_block'] = gjb

        if bool_temp is False:
            self.global_options['T_change'] = 0
        elif bool_temp is True:
            on_T = float(self._conf['change temperature']['change start'])
            off_T = float(self._conf['change temperature']['change finish'])
            rate_T = float(self._conf['change temperature']['change rate'])
            multi_T = float(self._conf['change temperature']['multiplier'])
            temper = [on_T, off_T, rate_T, multi_T]
            self.global_options['T_change'] = temper

        if bool_NaKblock is False:
            self.global_options['NaKATP_block'] = 0
        elif bool_NaKblock is True:
            on_nak = float(self._conf['block NaKATP pump']['change start'])
            off_nak = float(self._conf['block NaKATP pump']['change finish'])
            rate_nak = float(self._conf['block NaKATP pump']['change rate'])
            nak = [on_nak,off_nak,rate_nak]
            self.global_options['NaKATP_block'] = nak

        # Calcium TissueHandler: Calcium Induced Calcium Release (CICR).....................................................

        #FIXME: This is a bit unfortunate. Are full calcium dynamics
        #fundamentally broken? If not, perhaps we can define a new boolean in
        #our YAML configuration file to permit users to enable this. Endor!

        # Include full calcium dynamics in the situation (i.e. endoplasmic
        # reticulum, etc)?
        self.Ca_dyn = False

        #--------------------------------------------------------------------------------------------------------------
        #  CUSTOM GENERAL NETWORK (defined in main config file)
        #--------------------------------------------------------------------------------------------------------------

        self.network_config = self._conf.get('general network', None)

        if self.network_config is not None:
            self.molecules_enabled = self.network_config['implement network']
            # self.mol_mit_enabled = self.network_config['enable mitochondria']
            self.mol_mit_enabled = False
            self.expression_data_path_rel = self.network_config.get('expression data file', None)

            if (self.expression_data_path_rel is not None and
                self.expression_data_path_rel != 'None'):
                self.expression_data_path = pathnames.join(
                    self.conf_dirname, self.expression_data_path_rel)

                # Load this file under the assumption this file complies with a
                # sane version of the YAML specification.
                self.expression_data = yamls.load(
                    filename=self.expression_data_path,
                    yaml_version=YAML_VERSION)
        else:
            self.mol_mit_enabled = False

        #---------------------------------------------------------------------------------------------------------------
        #  GENE REGULATORY NETWORKS
        #---------------------------------------------------------------------------------------------------------------

        self.grn_enabled = self._conf['gene regulatory network settings']['gene regulatory network simulated']

        # If a GRN is enabled...
        if self.grn_enabled:
            # Load this GRN from this file under the assumption this file
            # complies with a sane version of the YAML specification.
            self.grn.load(
                conf_filename=self.grn_config_filename,
                yaml_version=YAML_VERSION)

        simgrndic = (
            self._conf['gene regulatory network settings']['sim-grn settings'])

        self.grn_dt = float(simgrndic.get('time step', 1.0e-2))
        self.grn_total_time = float(simgrndic.get('total time', 10.0))
        self.grn_tsample = float(simgrndic.get('sampling rate', 1.0))
        self.grn_runmodesim = simgrndic.get('run as sim', False)

        #--------------------------------------------------------------------------------------------------------------
        # VARIABLE SETTINGS
        #--------------------------------------------------------------------------------------------------------------

        self.T = float(self._conf['variable settings']['temperature'])  # system temperature

        # use the GHK equation to calculate alt Vmem from params?
        self.GHK_calc = self._conf['variable settings']['use Goldman calculator']

        # electroosmotic fluid flow-----------------------------------------------------
        # self.fluid_flow = self._conf['variable settings']['fluid flow']['include fluid flow']
        # self.mu_water = float(self._conf['variable settings']['fluid flow']['water viscocity']) # visc water [Pa.s]
        self.fluid_flow = False
        self.mu_water = 1.0e-2 # visc water [Pa.s]

        # electrodiffusive movement pumps and channels -----------------------------------
        self.sim_eosmosis = False
        # self.D_membrane = float(self._conf['variable settings']['channel electroosmosis']['membrane mobility'])
        # self.z_channel = float(self._conf['variable settings']['channel electroosmosis']['channel charge'])
        # self.z_pump = float(self._conf['variable settings']['channel electroosmosis']['pump charge'])

        # mechanical deformation ----------------------------------------------------------
        self.deformation = self._conf['variable settings']['deformation']['turn on']

        self.galvanotropism = float(self._conf['variable settings']['deformation']['galvanotropism'])
        self.td_deform = False # this has been disabled due to ongoing technical difficulties
        self.fixed_cluster_bound = self._conf['variable settings']['deformation']['fixed cluster boundary']
        self.youngMod = float(self._conf['variable settings']['deformation']['young modulus'])
        self.mu_tissue = float(self._conf['variable settings']['deformation']['viscous damping'])

        # osmotic and electrostatic pressures --------------------------------
        self.deform_osmo = self._conf['variable settings']['pressures']['include osmotic pressure']
        self.aquaporins = float(self._conf['variable settings']['pressures']['membrane water conductivity'])

        # calculate lame's parameters from young mod and the poisson ratio:
        self.poi = 0.49 # Poisson's ratio for the biological medium
        self.lame_mu = self.youngMod/(2*(1+self.poi))
        self.lame_lamb = (self.youngMod*self.poi)/((1+self.poi)*(1-2*self.poi))
        self.mu_membrane = 1.0 # membrane viscocity
        self.zeta = -70e-3  # zeta potential of cell membrane [V]

        # Gap junction parameters ------------------
        self.gj_surface = float(self._conf['variable settings']['gap junctions']['gap junction surface area'])
        self.gj_flux_sensitive = False
        self.gj_vthresh = float(self._conf['variable settings']['gap junctions']['gj voltage threshold'])
        self.gj_vgrad  = float(self._conf['variable settings']['gap junctions']['gj voltage window'])
        self.gj_min = float(self._conf['variable settings']['gap junctions']['gj minimum'])
        self.gj_respond_flow = False # (feature currently unsupported)
        self.v_sensitive_gj = self._conf['variable settings']['gap junctions']['voltage sensitive gj']

        # Microtubule properties........................................................................................

        # mtb = self.config['variable settings'].get('microtubules', {})

        # if mtb is not None:

        # self.use_microtubules = mtb.get('use microtubules', True)

        # self.use_microtubules = False
        # self.length_charge = float(mtb.get('charge per micrometer', -360.0))
        # self.mt_radius = float(mtb.get('radius', 15.0e-9))
        # self.mt_length = float(mtb.get('length', 5.0e-6))
        # self.tubulin_dipole = float(mtb.get('tubulin unit dipole', 1750))
        # self.tubulin_polar = float(mtb.get('tubulin polarizability', 50.0))
        #
        # self.tethered_tubule = (mtb.get('tethered tubule', True))
        #
        # self.cytoplasm_viscocity = 1.0e-2
        #
        # self.D_mtube = float(mtb.get('microtubule diffusion', 1.0))
        #
        # self.dilate_mtube_dt = float(mtb.get('time dilation factor', 1.0))
        #
        # self.mtube_init_x = mtb.get('initial x-coorinate', None)   # bitmap forcing alignment as initial state
        # self.mtube_init_y = mtb.get('initial y-coorinate', None)  # bitmap forcing alignment as initial state
        # self.mtube_init_rotangle = float(mtb.get('rotate initialization axis', 0.0))  # angle (in degrees) rotating axis
        #
        # if self.mtube_init_x == 'None':
        #     self.mtube_init_x = None
        #
        # if self.mtube_init_y == 'None':
        #     self.mtube_init_y = None
        #
        # if self.mtube_init_rotangle == 'None':
        #     self.mtube_init_rotangle = 0.0
        #
        # self.mt_space_density = mtb.get('microtubule density', None)
        #
        # if self.mt_space_density == 'None':
        #
        #     self.mt_space_density = None
        #
        # self.microtubules_orient_parallel = mtb.get('microtubules orient parallel', True)


        #...............................................................................................................

        # Environmental features and tight junctions ---------------------------------------------------

        #FIXME: This doesn't appear to be used anywhere, at the moment. Is this
        #safely removable now, or did we have Big Plans for this at some point?
        self.env_type = True # for now, can't handle air boundaries

        #FIXME: Should this actually be configurable? If not, no worries! -.-
        self.cluster_open = True

        self.D_tj = float(self._conf['variable settings']['tight junction scaling'])
        self.D_adh = float(self._conf['variable settings']['adherens junction scaling'])
        # tight junction relative ion movement properties:
        self.Dtj_rel = {}  # use a dictionary to hold the tj values:

        self.Dtj_rel['Na']=float(self._conf['variable settings']['tight junction relative diffusion']['Na'])
        self.Dtj_rel['K']=float(self._conf['variable settings']['tight junction relative diffusion']['K'])
        self.Dtj_rel['Cl']=float(self._conf['variable settings']['tight junction relative diffusion']['Cl'])
        self.Dtj_rel['Ca']=float(self._conf['variable settings']['tight junction relative diffusion']['Ca'])
        self.Dtj_rel['M']=float(self._conf['variable settings']['tight junction relative diffusion']['M'])
        self.Dtj_rel['P']=float(self._conf['variable settings']['tight junction relative diffusion']['P'])

        # environmental (global) boundary concentrations:
        self.cbnd = self._conf['variable settings']['env boundary concentrations']

        # include noise in the simulation?
        self.channel_noise_level = float(self._conf['variable settings']['noise']['static noise level'])

        self.dynamic_noise = self._conf['variable settings']['noise']['dynamic noise']
        self.dynamic_noise_level = float(self._conf['variable settings']['noise']['dynamic noise level'])

        # Modulator functions ------------------------------------------------------------------------------------------
        self.gradient_x_properties = {}
        self.gradient_y_properties = {}
        self.gradient_r_properties = {}

        self.periodic_properties = {}
        self.f_scan_properties = {}

        self.gradient_x_properties['slope'] =float(self._conf['modulator function properties']['gradient_x']['slope'])
        self.gradient_x_properties['x-offset'] =float(self._conf['modulator function properties']['gradient_x'].get('x-offset', 0.0))
        self.gradient_x_properties['z-offset'] =float(self._conf['modulator function properties']['gradient_x'].get('z-offset', 0.0))
        self.gradient_x_properties['exponent'] = float(
                                        self._conf['modulator function properties']['gradient_x'].get('exponent', 1))

        self.gradient_y_properties['slope'] =float(self._conf['modulator function properties']['gradient_y']['slope'])
        self.gradient_y_properties['x-offset'] = float(self._conf['modulator function properties']['gradient_y'].get('x-offset', 0.0))
        self.gradient_y_properties['z-offset'] = float(self._conf['modulator function properties']['gradient_y'].get('z-offset', 0.0))
        self.gradient_y_properties['exponent'] = float(
                                    self._conf['modulator function properties']['gradient_y'].get('exponent', 1))

        self.gradient_r_properties['slope'] = float(self._conf['modulator function properties']['gradient_r']['slope'])
        self.gradient_r_properties['x-offset'] = float(self._conf['modulator function properties']['gradient_r'].get('x-offset', 0.0))
        self.gradient_r_properties['z-offset'] = float(self._conf['modulator function properties']['gradient_r'].get('z-offset', 0.0))
        self.gradient_r_properties['exponent'] = float(self._conf['modulator function properties']['gradient_r'].get('exponent', 1))

        self.periodic_properties['frequency'] = float(self._conf['modulator function properties']['periodic']['frequency'])
        self.periodic_properties['phase'] = float(self._conf['modulator function properties']['periodic']['phase'])

        self.f_scan_properties['f start'] = float(
            self._conf['modulator function properties']['f_sweep']['start frequency'])
        self.f_scan_properties['f stop'] = float(
            self._conf['modulator function properties']['f_sweep']['end frequency'])

        #initialize the f vect field to None as it's set depending on the sim timestep:

        self.f_scan_properties['f slope'] = None

        # filename to read for bitmap loading gradient definition:
        chk = self._conf['modulator function properties'].get('gradient_bitmap', None)

        if chk is not None:
            self.grad_bm_fn = self._conf['modulator function properties']['gradient_bitmap']['file']
            self.grad_bm_offset = self._conf['modulator function properties']['gradient_bitmap'].get('z-offset', 0.0)
        else:
            self.grad_bm_fn = None
            self.grad_bm_offset = None

        chk2 = self._conf['modulator function properties'].get('single_cell', None)

        if chk2 is not None:
            self.mod_single_cell_offset = float(chk2['z-offset'])

        else:
            self.mod_single_cell_offset = None

        # ................{ EXPORTS ~ colormap                }................
        #FIXME: Non-ideal. Ideally, *ONLY* the name of each such colormap would
        #be stored in this object. Storing the actual colormap object as we do
        #here requires us to also pickle these actual colormap objects with
        #every seed, initialization, and simulation. (Which is bad.)
        self.default_cm = mplcolormap.get_colormap(
            self.colormap_diverging_name)
        self.background_cm = mplcolormap.get_colormap(
            self.colormap_sequential_name)
        self.gj_cm = mplcolormap.get_colormap(
            self.colormap_gj_name)
        self.network_cm = mplcolormap.get_colormap(
            self.colormap_grn_name)

        # ................{ EXPORTS ~ plot                    }................
        ro = self._conf['results options']

        #FIXME: Replace all instances of "p.turn_all_plots_off" in the codebase
        #by "not p.plot.is_after_sim_show" and remove this attribute entirely.
        self.turn_all_plots_off = not self.plot.is_after_sim_show

        #FIXME: Replace all instances of "p.autosave" in the codebase
        #by "p.plot.is_after_sim_save" and remove this attribute entirely.
        self.autosave = self.plot.is_after_sim_save  # autosave all still images to a results directory

        self.plot_cutlines = ro['plot cutlines']

        # new options for plotting reaction network graphs:
        self.plot_network = ro['plot networks']

        # Colors.
        self.vcolor = ro['vector and stream color']  # color of vector and streamlines

        self.plot_networks_single_cell = ro['plot networks single cell']
        self.showCells = ro['show cells']     # True = polygon patch plots, False = trimesh
        self.stream_density = ro['streamline density']
        self.IecmPlot = ro['plot total current']    # True = plot extracellular currents, false plot gj
        self.plotMask = ro['plot masked geometry']

        #FIXME: Remove all of the following after globally removing "plot seed".
        # Plot seed options.
        self.plot_cluster_mask = ro.get('plot cluster mask', True)

        #--------------------------------------------------------------------------------------------------------------
        # INTERNAL USE ONLY
        #--------------------------------------------------------------------------------------------------------------

        iu = self._conf['internal parameters']

        self.interp_type = 'nearest'

        # self.bound_cell_clip_ratio = iu.get('boundary cell size cutoff', 0.5)
        self.substances_affect_charge = iu['substances affect Vmem']  # Do Network substances function bioelectrically?

        # default free diffusion constants (cytoplasmic)
        self.Do_Na = float(iu['Do_Na'])    # free diffusion constant sodium [m2/s]
        self.Do_K  = float(iu['Do_K'])     # free diffusion constant potassium [m2/s]
        self.Do_Cl = float(iu['Do_Cl'])    # free diffusion constant chloride [m2/s]
        self.Do_Ca = float(iu['Do_Ca'])    # free diffusion constant calcium [m2/s]
        self.Do_M  = float(iu['Do_M'])     # free diffusion constant mystery anchor ion [m2/s]
        self.Do_P  = float(iu['Do_P'])     # free diffusion constant protein [m2/s]

        # fixed levels of ATP, ADP and Pi, required for use in pump equations to get propper kinetics
        self.cATP = 1.5
        self.cADP = 0.1
        self.cPi = 0.1

        # pump parameters
        self.alpha_NaK = float(iu['alpha_NaK']) # maximum rate constant sodium-potassium ATPase per unit surface area

        # FIXME add these as options to the config
        self.KmNK_Na = 12.0   # NaKATPase enzyme ext Na half-max sat value (alpha1 = 12, alpha2 = 20, alpha3 = 60)
        self.KmNK_K = 0.2     # NaKATPase enzyme ext K half-max sat value (alpha1 = 0.2, alpha2 = 0.20, alpha3 = 0.09)
        self.KmNK_ATP = 0.5   # NaKATPase enzyme ATP half-max sat value

        self.alpha_Ca = float(iu['alpha_Ca']) # pump rate for calcium ATPase in membrane [1/mol*s] 2.0e-15

        # FIXME add these as options to the config:
        self.KmCa_Ca = 1.0e-3   # CaATPase enzyme Ca half-max sat value (1.7 - 2.8 for vascular, 0.25 for platlets)
        self.KmCa_ATP = 0.5    # CaATPase enzyme ATP half-max sat value

        # partial pressure dissolved CO2
        self.CO2 = 40.0   # [mmHg]
        self.cCO2 = 1.2

        # charge states of ions
        self.z_Na = 1
        self.z_K = 1
        self.z_Cl = -1
        self.z_Ca = 2
        self.z_H = 1
        self.z_P = -1
        self.z_M = -1

        # molar mass of ions (kg/mol):
        self.M_Na = 23e-3
        self.M_K = 39e-3
        self.M_Cl = 35e-3
        self.M_Ca = 40e-3
        self.M_H = 1e-3
        self.M_P = 500e-3
        self.M_M = 60e-3

        # fundamental constants
        self.F = 96485 # Faraday constant [J/V*mol]
        self.R = 8.314  # Gas constant [J/K*mol]
        self.eo = 8.854e-12 # permittivity of free space [F/m]
        self.kb = 1.3806e-23  # Boltzmann constant [m2 kg/ s2 K1]
        self.q = 1.602e-19    # electron charge [C]
        self.mu = 1.275e-6   # magnetic permeability [H/m or N/A2]
        self.NAv = 6.022e23     # Avagadro's Number
        self.er = 80.0          # relative dielectric constant of electrolyte
        self.eedl = float(iu.get('dielectric constant', 40.0))         # relative dielectric constant of screening layer

        self.deltaGATP = -37000    # free energy released in ATP hydrolysis under standard phys conditions [J/mol]

        self.ac = 1.0e-6  # cell-cell separation for drawing
        self.scale_cell = 0.99          # the amount to scale cell membranes in from ecm edges (only affects drawing)
        self.cm = float(iu['membrane capacitance'])           # patch capacitance of cell membrane 0.022 [F/m2]

        self.tm = 7.5e-9           # thickness of cell membrane [m]
        self.cell_sides = 4      # minimum number of membrane domains per cell (must be >2)
        self.scale_alpha = 1.4   # the amount to scale (1/d_cell) when calculating the concave hull (boundary search)
        self.merge_cut_off = (1/50)  # the fraction of nominal cell perimeter at which nearby ecm points are merged

        self.d_cell = self.cell_radius * 2  # diameter of single cell
        self.nx = int(self.wsx / self.d_cell)  # number of lattice sites in world x index
        self.ny = int(self.wsy / self.d_cell)  # number of lattice sites in world y index
        self.wsx = self.wsx + 5 * self.cell_lattice_disorder * self.d_cell  # readjust the world size for noise
        self.wsy = self.wsy + 5 * self.cell_lattice_disorder * self.d_cell

        self.gjl = 2*self.tm + self.cell_space     # gap junction length
        self.gj_radius = 1.0e-9              # effective radius of gap junctions connecting cells [m] (range 0 to 5.0 e-9 m)

        self.um = 1e6    # multiplication factor to convert m to um

        self.isamples = 40.0  # sampling of vector data for currents

        self.rho = 1050 # mass density of system [kg/m3]

        self.fast_update_ecm = iu.get('fast update ecm', False)  # quick or slow update to cell<--> ecm grid exchange?

        self.sharpness = float(iu.get('sharpness env', 0.999))

        self.smooth_cells = 1/float(iu.get('sharpness cell', 0.5))

        self.true_cell_size = float(iu.get('true cell size', 1.0e-5))

        #FIXME: Can this initialization be safely moved earlier -- say, directly
        #*AFTER* tissue profile initialization required by this initialization?

        # Initialize the ion profile specified by this configuration.
        self._load_ion_profile()


    #FIXME: Ideally, this method should be private. Unfortunately, external
    #callers (notably, tests) currently need to call this method to update
    #absolute pathnames after modifying relative pathnames. The solution, of
    #course, is to have Python implicitly call this method whenever *ANY*
    #relative pathname internally referenced by this class is modified. Doing
    #so correctly will require augmenting the expr_alias() data descriptor to
    #support yet-another-optional-keyword-parameter enabling callers to pass a
    #callable to be implicitly called whenever that descriptor is set -- say,
    #"callback_set" or "on_set". Given that, we would then refactor the above
    #YAML aliases to resemble something like:
    #
    #    sim_pickle_basename = yaml_alias(
    #        "['sim file saving']['file']", str, on_set=self.reload_paths)
    #
    #Oh... wait. Obviously, we have no access to self.reload_paths() from class
    #scope. Err; perhaps refactor that to require a method accepting the
    #current "Parameters" object be passed, which would then permit convenient
    #calling via lambdas ala:
    #
    #    sim_pickle_basename = yaml_alias(
    #        "['sim file saving']['file']", str,
    #        on_set: lambda p: p.reload_paths())
    #
    #Right. That's definitely it. Given that, we could then define a
    #yaml_alias() derivative specific to simulation paths in this submodule
    #resembling:
    #
    #    from betse.util.type.call import callables
    #    _yaml_alias_path_absolute = callables.make_partial(
    #        func=yaml_alias,
    #        kwargs={
    #            'cls': str,
    #            'callback_set': lambda p: p.reload_paths(),
    #        },
    #        doc='''...''',
    #    )
    #
    #Works for us. *shrug*
    #FIXME: After doing so:
    #
    #* Remove *ALL* external calls to this method.
    #* Rename this method to _load_paths() for orthogonality.
    def reload_paths(self) -> None:
        '''
        Redefine all absolute pathnames depending upon relative pathnames
        specified by this configuration.

        To avoid desynchronization between the two, this method *must* be
        called on every change to such a relative pathname.
        '''

        # Absolute dirnames to which phase results are pickled.
        self.init_pickle_dirname = pathnames.join_and_canonicalize(
            self.conf_dirname, self.init_pickle_dirname_relative)
        self.sim_pickle_dirname = pathnames.join_and_canonicalize(
            self.conf_dirname, self.sim_pickle_dirname_relative)

        # Absolute filenames to which phase results are pickled, defined after
        # absolute dirnames to these results.
        self.init_pickle_filename = pathnames.join(
            self.init_pickle_dirname, self.init_pickle_basename)
        self.seed_pickle_filename = pathnames.join(
            self.init_pickle_dirname, self.seed_pickle_basename)
        self.sim_pickle_filename = pathnames.join(
            self.sim_pickle_dirname, self.sim_pickle_basename)

        # Absolute dirnames to which phase results are exported.
        self.init_export_dirname = pathnames.join_and_canonicalize(
            self.conf_dirname, self.init_export_dirname_relative)
        self.sim_export_dirname = pathnames.join_and_canonicalize(
            self.conf_dirname, self.sim_export_dirname_relative)

        #FIXME: Functional test this up, which was silently broken for a
        #shamefully lengthy period of time. To do so, we'll want to test
        #from a different directory other than the directory containing
        #the top-level YAML file.
        #FIXME: Generalize "grn_config_filename_relative" into a YAML alias.
        #FIXME: Reduce "grn_config_*" to merely "grn_conf_*".

        # Relative filename from which the GRN is configured.
        self.grn_config_filename_relative = (
            self._conf['gene regulatory network settings']['gene regulatory network config'])

        # Absolute filename from which the GRN is configured.
        self.grn_config_filename = pathnames.join(
            self.conf_dirname, self.grn_config_filename_relative)

        # Absolute pathnames to which GRN results are pickled.
        self.grn_pickle_dirname = pathnames.join(
            self.conf_dirname, self.grn_pickle_dirname_relative)
        self.grn_pickle_filename = pathnames.join(
            self.grn_pickle_dirname, self.grn_pickle_basename)

        # Absolute pathnames from which prior GRN results are unpickled.
        self.grn_unpickle_filename = None
        if self.grn_unpickle_filename_relative is not None:
            # logs.log_debug('GRN load from: %s', self.grn_unpickle_filename_relative)
            self.grn_unpickle_filename = pathnames.join(
                self.conf_dirname, self.grn_unpickle_filename_relative)

        # Create all non-existing directories.
        dirs.make_unless_dir(
            self.init_pickle_dirname, self.sim_pickle_dirname,
            self.init_export_dirname, self.sim_export_dirname,
            self.grn_pickle_dirname,
        )


    def _load_ion_profile(self) -> None:
        '''
        Initialize the ion profile specified by this configuration.
        '''

        # Default tissue profile providing base membrane diffusion constants.
        base = self.tissue_default

        # simplest ion profile giving realistic results with minimal ions (Na+ & K+ focus):
        if self.ion_profile is IonProfileType.BASIC:
            self.cNa_env = 145.0
            self.cK_env = 5.0
            self.cP_env = 10.0

            zs = [self.z_Na, self.z_K, self.z_P]

            conc_env = [self.cNa_env, self.cK_env, self.cP_env]
            self.conc_env_m, self.z_M_env = _balance_charge(conc_env,zs)

            assert self.z_M_env == -1

            self.cNa_cell = 12.0
            self.cK_cell = 139.0
            self.cP_cell = 135.0

            conc_cell = [self.cNa_cell,self.cK_cell, self.cP_cell]

            self.cM_cell, self.z_M_cell = _balance_charge(conc_cell,zs)

            assert self.z_M_cell == -1

            self.ions_dict = {'Na':1, 'K':1, 'Cl':0, 'Ca':0, 'H':0, 'P':1, 'M':1}

            self.cell_concs ={'Na': self.cNa_cell, 'K': self.cK_cell, 'P': self.cP_cell, 'M': self.cM_cell}
            self.env_concs ={'Na': self.cNa_env, 'K': self.cK_env, 'P': self.cP_env, 'M': self.conc_env_m}
            self.mem_perms = {'Na': base.Dm_Na, 'K': base.Dm_K, 'P': base.Dm_P, 'M': base.Dm_M}
            self.ion_charge = {'Na': self.z_Na, 'K': self.z_K, 'P': self.z_P, 'M': self.z_M}
            self.free_diff = {'Na': self.Do_Na, 'K': self.Do_K, 'P': self.Do_P, 'M': self.Do_M}
            self.molar_mass = {'Na': self.M_Na, 'K': self.M_K, 'P': self.M_P, 'M': self.M_M}
            self.ion_long_name = {'Na':'sodium', 'K':'potassium', 'P':'proteins', 'M':'anion'}

        elif self.ion_profile is IonProfileType.BASIC_CA:
            self.cNa_env = 145.0
            self.cK_env = 5.0
            self.cCa_env = 2.0
            self.cP_env = 10.0

            zs = [self.z_Na, self.z_K, self.z_Ca, self.z_P]

            conc_env = [self.cNa_env, self.cK_env, self.cCa_env, self.cP_env]
            self.conc_env_m, self.z_M_env = _balance_charge(conc_env,zs)

            assert self.z_M_env == -1

            self.cNa_cell = 12.0
            self.cK_cell = 139.0
            self.cCa_cell = 1.0e-4
            self.cP_cell = 135.0

            conc_cell = [self.cNa_cell,self.cK_cell, self.cCa_cell, self.cP_cell]

            self.cM_cell, self.z_M_cell = _balance_charge(conc_cell,zs)

            assert self.z_M_cell == -1

            self.cCa_er = 0.5
            self.cM_er = self.cCa_er

            self.ions_dict = {'Na':1, 'K':1, 'Cl':0, 'Ca':1, 'H':0, 'P':1, 'M':1}

            self.cell_concs ={'Na': self.cNa_cell, 'K': self.cK_cell, 'Ca': self.cCa_cell, 'P': self.cP_cell, 'M': self.cM_cell}
            self.env_concs ={'Na': self.cNa_env, 'K': self.cK_env, 'Ca': self.cCa_env, 'P': self.cP_env, 'M': self.conc_env_m}
            self.mem_perms = {'Na': base.Dm_Na, 'K': base.Dm_K, 'Ca': base.Dm_Ca, 'P': base.Dm_P, 'M': base.Dm_M}
            self.ion_charge = {'Na': self.z_Na, 'K': self.z_K, 'Ca': self.z_Ca, 'P': self.z_P, 'M': self.z_M}
            self.free_diff = {'Na': self.Do_Na, 'K': self.Do_K, 'Ca': self.Do_Ca, 'P': self.Do_P, 'M': self.Do_M}
            self.molar_mass = {'Na': self.M_Na, 'K': self.M_K, 'Ca': self.M_Ca, 'P': self.M_P, 'M': self.M_M}
            self.ion_long_name = {'Na':'sodium', 'K':'potassium', 'Ca':'calcium', 'P':'proteins', 'M':'anion'}

        # default environmental and cytoplasmic initial values mammalian cells
        elif self.ion_profile is IonProfileType.MAMMAL:
            # initialize proton concentrations to "None" placeholders
            self.cNa_env = 145.0
            self.cK_env = 5.0
            self.cCl_env = 115.0
            self.cCa_env = 2.0
            self.cP_env = 10.0

            zs = [self.z_Na, self.z_K, self.z_Cl, self.z_Ca, self.z_P]

            conc_env = [self.cNa_env, self.cK_env, self.cCl_env, self.cCa_env, self.cP_env]
            self.conc_env_m, self.z_M_env = _balance_charge(conc_env,zs)

            assert self.z_M_env == -1

            self.cNa_cell = 12.0
            self.cK_cell = 139.0
            self.cCl_cell = 4.0
            self.cCa_cell = 5.0e-5
            self.cP_cell = 135.0

            conc_cell = [self.cNa_cell,self.cK_cell, self.cCl_cell, self.cCa_cell, self.cP_cell]
            self.cM_cell, self.z_M_cell = _balance_charge(conc_cell,zs)

            assert self.z_M_cell == -1

            self.cCa_er = 0.5
            self.cM_er = self.cCa_er

            self.ions_dict = {'Na':1, 'K':1, 'Cl':1, 'Ca':1, 'H':0, 'P':1, 'M':1}

            self.cell_concs ={'Na': self.cNa_cell, 'K': self.cK_cell, 'Ca': self.cCa_cell, 'Cl': self.cCl_cell,
                              'P': self.cP_cell, 'M': self.cM_cell}

            self.env_concs ={'Na': self.cNa_env, 'K': self.cK_env, 'Ca': self.cCa_env, 'Cl': self.cCl_env,
                             'P': self.cP_env, 'M': self.conc_env_m}

            self.mem_perms = {'Na': base.Dm_Na, 'K': base.Dm_K, 'Ca': base.Dm_Ca, 'Cl': base.Dm_Cl,
                              'P': base.Dm_P, 'M': base.Dm_M}

            self.ion_charge = {'Na': self.z_Na, 'K': self.z_K, 'Ca': self.z_Ca, 'Cl': self.z_Cl,
                               'P': self.z_P, 'M': self.z_M}

            self.free_diff = {'Na': self.Do_Na, 'K': self.Do_K, 'Ca': self.Do_Ca, 'Cl': self.Do_Cl,
                              'P': self.Do_P, 'M': self.Do_M}

            self.molar_mass = {'Na': self.M_Na, 'K': self.M_K, 'Ca': self.M_Ca, 'Cl': self.M_Cl,
                               'P': self.M_P, 'M': self.M_M}

            self.ion_long_name = {'Na':'sodium', 'K':'potassium', 'Ca':'calcium', 'Cl':'chloride',
                                  'P':'proteins', 'M':'anion'}

        elif self.ion_profile is IonProfileType.AMPHIBIAN:
            # initialize proton concentrations to "None" placeholders
            self.cNa_env = 14.50
            self.cK_env = 0.5
            self.cCl_env = 10.50
            self.cCa_env = 0.2
            self.cP_env = 0.0

            zs = [self.z_Na, self.z_K, self.z_Cl, self.z_Ca, self.z_P]

            conc_env = [self.cNa_env, self.cK_env, self.cCl_env, self.cCa_env, self.cP_env]
            self.conc_env_m, self.z_M_env = _balance_charge(conc_env,zs)

            assert self.z_M_env == -1

            self.cNa_cell = 8.0
            self.cK_cell = 125.0
            self.cCl_cell = 20.0
            self.cCa_cell = 1.0e-3
            self.cP_cell = 100.0

            conc_cell = [self.cNa_cell,self.cK_cell, self.cCl_cell, self.cCa_cell, self.cP_cell]
            self.cM_cell, self.z_M_cell = _balance_charge(conc_cell,zs)

            assert self.z_M_cell == -1

            self.cCa_er = 0.5
            self.cM_er = self.cCa_er

            self.ions_dict = {'Na':1, 'K':1, 'Cl':1, 'Ca':1, 'H':0, 'P':1, 'M':1}

            self.cell_concs ={'Na': self.cNa_cell, 'K': self.cK_cell, 'Ca': self.cCa_cell, 'Cl': self.cCl_cell,
                              'P': self.cP_cell, 'M': self.cM_cell}
            self.env_concs ={'Na': self.cNa_env, 'K': self.cK_env, 'Ca': self.cCa_env, 'Cl': self.cCl_env,
                             'P': self.cP_env, 'M': self.conc_env_m}
            self.mem_perms = {'Na': base.Dm_Na, 'K': base.Dm_K, 'Ca': base.Dm_Ca, 'Cl': base.Dm_Cl,
                              'P': base.Dm_P, 'M': base.Dm_M}
            self.ion_charge = {'Na': self.z_Na, 'K': self.z_K, 'Ca': self.z_Ca, 'Cl': self.z_Cl,
                               'P': self.z_P, 'M': self.z_M}
            self.free_diff = {'Na': self.Do_Na, 'K': self.Do_K, 'Ca': self.Do_Ca, 'Cl': self.Do_Cl,
                              'P': self.Do_P, 'M': self.Do_M}

            self.molar_mass = {'Na': self.M_Na, 'K': self.M_K, 'Ca': self.M_Ca, 'Cl': self.M_Cl,
                               'P': self.M_P, 'M': self.M_M}

            self.ion_long_name = {'Na':'sodium', 'K':'potassium', 'Ca':'calcium', 'Cl':'chloride',
                                  'P':'proteins', 'M':'anion'}

        # user-specified environmental and cytoplasm values (customized)
        elif self.ion_profile is IonProfileType.CUSTOM:
            # initialize proton concentrations to "None" placeholders
            cip = self._conf['general options']['customized ion profile']

            self.cNa_env = float(cip['extracellular Na+ concentration'])
            self.cK_env = float(cip['extracellular K+ concentration'])
            self.cCl_env = float(cip['extracellular Cl- concentration'])
            self.cCa_env = float(cip['extracellular Ca2+ concentration'])
            self.cP_env = float(cip['extracellular protein- concentration'])

            zs = [self.z_Na, self.z_K, self.z_Cl, self.z_Ca, self.z_P]

            conc_env = [self.cNa_env, self.cK_env, self.cCl_env, self.cCa_env, self.cP_env]
            self.conc_env_m, self.z_M_env = _balance_charge(conc_env, zs)

            self.cNa_cell = float(cip['cytosolic Na+ concentration'])
            self.cK_cell = float(cip['cytosolic K+ concentration'])
            self.cCl_cell = float(cip['cytosolic Cl- concentration'])
            self.cCa_cell = float(cip['cytosolic Ca2+ concentration'])
            # self.cM_cell = float(cip['cytosolic HCO3- concentration'])
            self.cP_cell = float(cip['cytosolic protein- concentration'])

            conc_cell = [self.cNa_cell, self.cK_cell, self.cCl_cell, self.cCa_cell, self.cP_cell]
            self.cM_cell, self.z_M_cell = _balance_charge(conc_cell, zs)

            assert self.z_M_cell == -1

            # self.cCa_er = float(cip['endoplasmic reticulum Ca2+'])
            self.cCa_er = 0.1
            self.cM_er = self.cCa_er

            self.ions_dict = {'Na':1, 'K':1, 'Cl':1, 'Ca':1, 'H':0, 'P':1, 'M':1}

            self.cell_concs ={'Na': self.cNa_cell, 'K': self.cK_cell, 'Ca': self.cCa_cell, 'Cl': self.cCl_cell, 'P': self.cP_cell, 'M': self.cM_cell}

            self.env_concs ={'Na': self.cNa_env, 'K': self.cK_env, 'Ca': self.cCa_env, 'Cl': self.cCl_env, 'P': self.cP_env, 'M': self.conc_env_m}

            self.mem_perms = {'Na': base.Dm_Na, 'K': base.Dm_K, 'Ca': base.Dm_Ca, 'Cl': base.Dm_Cl, 'P': base.Dm_P, 'M': base.Dm_M}

            self.ion_charge = {'Na': self.z_Na, 'K': self.z_K, 'Ca': self.z_Ca, 'Cl': self.z_Cl, 'P': self.z_P, 'M': self.z_M}

            self.free_diff = {'Na': self.Do_Na, 'K': self.Do_K, 'Ca': self.Do_Ca, 'Cl': self.Do_Cl, 'P': self.Do_P, 'M': self.Do_M}

            self.molar_mass = {'Na': self.M_Na, 'K': self.M_K, 'Ca': self.M_Ca, 'Cl': self.M_Cl, 'P': self.M_P, 'M': self.M_M}

            self.ion_long_name = {'Na':'sodium', 'K':'potassium', 'Ca':'calcium', 'Cl':'chloride',
                                  'P':'proteins', 'M':'anion'}

        # Else, this ion profile type is unrecognized. Raise an exception.
        else:
            raise BetseSimConfException(
                'Ion profile type "{}" unrecognized.'.format(self.ion_profile))

    # ..................{ UNLOADERS                         }..................
    def unload(self) -> None:

        # Unload our superclass.
        super().unload()

        # Unload all pathname-specific instance variables.
        self._unload_paths()

        # Unload all previously loaded network subconfigurations.
        self.grn.unload()

        # Unload all previously loaded tissue subconfigurations.
        self.cut_profiles   .unload()
        self.tissue_default .unload()
        self.tissue_profiles.unload()

        # Unload all previously loaded export subconfigurations.
        self.anim  .unload()
        self.csv   .unload()
        self.plot  .unload()
        self.visual.unload()


    def _unload_paths(self) -> None:
        '''
        Deassociate the pathname-specific subset of this configuration.

        Specifically, this method nullifies all absolute pathnames depending
        upon relative pathnames for safety.
        '''

        self.grn_pickle_dirname = None
        self.grn_pickle_filename = None
        self.grn_unpickle_filename = None
        self.init_export_dirname = None
        self.init_pickle_dirname = None
        self.init_pickle_filename = None
        self.seed_pickle_filename = None
        self.sim_export_dirname = None
        self.sim_pickle_dirname = None
        self.sim_pickle_filename = None

    # ..................{ EXCEPTIONS                        }..................
    def die_unless_ecm(self) -> None:
        '''
        Raise an exception unless this configuration has enabled simulation of
        the extracellular matrix and hence environmental grid spaces.
        '''

        if not self.is_ecm:
            raise BetseSimConfException(
                'Extracellular spaces disabled by '
                'this simulation configuration.')

    # ..................{ SUPERCLASS                        }..................
    def _iter_conf_subdir_basenames(self) -> IterableTypes:

        # If no YAML file has been loaded yet, raise an exception.
        self.die_unless_loaded()

        # Absolute dirname of the directory containing the default
        # YAML-formatted simulation configuration file.
        DATA_YAML_DIRNAME = appmetaone.get_app_meta().betse_data_yaml_dirname

        # Default to the basenames of all direct subdirectories of the
        # directory containing this file, which are guaranteed to be required
        # by this file.
        return dirs.iter_subdir_basenames(DATA_YAML_DIRNAME)

# ....................{ HELPERS                           }....................
def _balance_charge(concentrations: SequenceTypes, zs: SequenceTypes) -> tuple:
    '''
    Sum the concentrations of profile ions with their charge state to
    determine how much net positive charge exists, returning the concentration
    of the charge compensation anion M- needed to have zero net charge.

    Parameters
    -------------
    concentrations : SequenceType
        Array defining the concentrations of all ions in a space.
    zs : SequenceType
        Array (in complementary order to `concentrations`) of ion valence
        state.

    Returns
    ---------
    (float, float)
        2-tuple `(bal_conc, valence)`, where:

        * ``bal_conc`` is the concentration of anion M- to create zero net
          charge.
        * ``valence`` is the charge of the ``bal_conc``. Ideally, this should
          *always* be -1.
    '''

    q = 0

    for conc,z in zip(concentrations, zs):
        q = q+ conc*z

        to_zero = -q
        bal_conc = abs(to_zero)
        valance = np.sign(to_zero)

        assert bal_conc >= 0

    return bal_conc, valance



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **application pathname** (i.e., pathnames relative to the current
application, regardless of the package manager used to install that application
on the local filesystem) hierarchy.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from modules guaranteed to exist at the start of
# installation. This includes all standard Python and application modules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation. Likewise, to avoid circular
# import dependencies, the top-level of this module should avoid importing
# application modules where feasible.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# from betse.util.io.log import logs
from betse.util.type.types import type_check, ModuleOrStrTypes

# ....................{ GETTERS                            }....................
@type_check
def get_pathname(package: ModuleOrStrTypes, pathname: str) -> str:
    '''
    Absolute pathname canonicalized from the passed relative pathname in a
    portable manner guaranteed to be relative to the absolute dirname of the
    passed application package if a path with this canonical pathname exists
    *or* raise an exception otherwise (i.e., if no such path exists).

    Specifically, this method returns:

    * If this application is a PyInstaller-frozen executable binary, the
      concatenation of (in order):

      #. The absolute path of the temporary directory containing all
         application data resources extracted from this binary by this
         executable's bootloader as specified by the PyInstaller-specific
         private attribute ``_MEIPASS`` injected into the canonical :mod:`sys`
         module by the PyInstaller bootloader embedded in this binary. "And
         it's turtles all the way down."
      #. The passed relative pathname.

    * If this application is a :mod:`setuptools`-installed script wrapper, the
      result of querying :mod:`setuptools` for the absolute path of the passed
      relative pathname. In this case, this path will have been preserved as is
      in the :mod:`setuptools`-installed copy of this application in the
      package tree for the active Python interpreter.
    * Else, the concatenation of (in order):

      #. The absolute path of the directory providing this root package.
      #. The passed relative pathname.

      In this case, this application is typically either a
      :mod:`setuptools`-symlinked script wrapper *or* was invoked via the
      secretive ``python3 -m {package.__name__}`` command.

    Parameters
    ----------
    package : ModuleOrStrTypes
        Topmost package of the application to canonicalize this pathname for,
        defined as either:

        * The fully-qualified name of this package, in which case this function
          dynamically imports this package.
        * A previously imported package object.
    pathname : str
        Relative pathname of the path to be canonicalized.

    Returns
    ----------
    Absolute pathname of this path relative to the absolute pathname of this
    application package.

    Raises
    ----------
    BetseModuleException
        If this package is a subpackage rather than topmost.
    BetsePathException
        If no path with the absolute pathname to be returned exists.
    BetsePathnameException
        If this pathname is absolute rather than relative.
    '''

    # Avoid circular import dependencies.
    # from betse.lib.setuptools import supresource
    from betse.util.path import pathnames, paths
    from betse.util.py import pyfreeze
    from betse.util.py.module import pymodule

    # If this package is *NOT* topmost, raise an exception.
    pymodule.die_unless_topmost(package)

    # If this pathname is absolute rather than relative, raise an exception.
    pathnames.die_if_absolute(pathname)

    # # Name of this package.
    # package_name = pymodule.get_name_qualified(package)

    # If this application is frozen by PyInstaller, canonicalize this path
    # relative to the directory to which this application is unfrozen.
    if pyfreeze.is_frozen_pyinstaller():
        # Absolute dirname of the directory containing this frozen application.
        app_frozen_dirname = pyfreeze.get_app_dirname_pyinstaller()

        #FIXME: This requires generalization to the passed package, if in fact
        #this can be generalized. If this cannot be generalized, then some
        #other means will be needed to achieve the same or a similar effect.
        app_pathname = pathnames.join(app_frozen_dirname, pathname)
    #FIXME: It's unclear exactly what the Hatch-specific equivalent to a
    #setuptools-specific "resource" is or even if there *IS* an equivalent. For
    #now, let's deprecate all "supresource" usage entirely. *sigh*
    # # Else if this application is a setuptools-installed script wrapper,
    # # canonicalize this path by deferring to the setuptools resource API.
    # elif supresource.is_dir(
    #     module_name=package_name, dirname=pathname):
    #     app_pathname = supresource.get_pathname(
    #         module_name=package_name, pathname=pathname)
    # Else, the current application is either a setuptools-symlinked script
    # wrapper *OR* was invoked via the secretive "python3 -m betse"
    # command. In either case, this directory's path is directly obtainable
    # relative to the absolute path of the passed package.
    else:
        # Absolute canonical dirname of the directory defining this package.
        package_dirname = pymodule.get_dirname_canonical(package)

        # Canonicalize this path relative to this directory.
        app_pathname = pathnames.join(package_dirname, pathname)

    # If this path is not found, fail; else, return this path.
    return paths.path_or_die(app_pathname)


@type_check
def get_dirname(package: ModuleOrStrTypes, dirname: str) -> str:
    '''
    Absolute dirname canonicalized from the passed relative dirname in a
    portable manner guaranteed to be relative to the absolute dirname of the
    passed application package if a directory with this canonical dirname
    exists *or* raise an exception otherwise (i.e., if no such directory
    exists).

    Raises
    ----------
    BetseDirException
        If no directory with the absolute dirname to be returned exists.

    See Also
    ----------
    :func:`get_pathname`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs

    # Absolute dirname of this directory.
    app_dirname = get_pathname(package=package, pathname=dirname)

    # If this directory is not found, fail; else, return this directory.
    return dirs.dir_or_die(app_dirname)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Yet Another Markup Language (YAML) **representer** (i.e., callable serializing
all objects of the same type into well-formatted YAML) functionality.
'''

# ....................{ IMPORTS                            }....................
# from betse.util.type.types import type_check  #, MappingOrSequenceTypes

# ....................{ GLOBALS ~ private                  }....................
_TAG_NUMPY_NDARRAY = '!ndarray'
'''
YAML-formatted tag identifying each YAML sequence converted from a Numpy array.
'''

# ....................{ ADDERS                             }....................
def add_representers(representer: object) -> None:
    '''
    Add all **custom representers** (i.e., callables serializing all objects of
    the same type into well-formatted YAML) required by this application to the
    passed :class:`yaml.representer.Representer`-like object.

    Motivation
    ----------
    This function ensures that dumping arbitrary objects to YAML with the
    passed dumper will stringify *all*:

    * Numpy arrays as YAML sequences.
    * Numpy booleans as YAML booleans.
    * Numpy complex numbers as YAML complex numbers.
    * Numpy floats as YAML floats.
    * Numpy integers as YAML integers.

    Since most YAML implementations lack explicit support for Numpy types,
    attempting to saving Numpy objects to YAML results in non-human-readable
    output defeating the whole purpose of YAML. This output resembles:

        >>> print(yaml.dump({'a':numpy.array([3.1415,])}))
        /usr/lib64/python3.4/site-packages/yaml/representer.py:135:
        FutureWarning: comparison to `None` will result in an elementwise
        object comparison in the future.
          if data in [None, ()]:
        a: !!python/object/apply:numpy._core.multiarray._reconstruct
          args:
          - !!python/name:numpy.ndarray ''
          - !!python/tuple [0]
          - !!binary |
            Yg==
          state: !!python/tuple
          - 1
          - !!python/tuple [1]
          - !!python/object/apply:numpy.dtype
            args: [f8, 0, 1]
            state: !!python/tuple [3, <, null, null, null, -1, -1, 0]
          - false
          - !!binary |
            bxKDwMohCUA=

    Parameters
    ----------
    representer: object
        :class:`yaml.representer.Representer`-like object converting arbitrary
        Python objects to YAML-formatted strings.  This object *must* define an
        ``add_representer`` callable accepting the type to be represented and a
        callable converting objects of that type to YAML strings. This object
        may otherwise be of any type (e.g., module).
    '''

    # Defer third-party imports.
    import numpy

    # Represent all core Numpy types.
    representer.add_representer(numpy.ndarray,    _represent_numpy_ndarray)
    representer.add_representer(numpy.bool,       _represent_numpy_bool)
    representer.add_representer(numpy.complex128, _represent_numpy_complex)
    representer.add_representer(numpy.float64,    _represent_numpy_float)
    representer.add_representer(numpy.int64,      _represent_numpy_int)

# ....................{ REPRESENTERS                       }....................
def _represent_numpy_ndarray(dumper, ndarray: 'numpy.ndarray') -> str:
    '''
    Convert the passed Numpy array into a YAML-formatted string.

    Specifically, this function returns a stringified YAML mapping whose:

    * Keys are the names of parameters accepted by the :func:`numpy.array`
      function.
    * Values are the values of these parameters required to losslessly restore
      the contents of this array on YAML deserialization of this string.

    Parameters
    ----------
    dumper: yaml.Dumper
        Object converting arbitrary Python objects to YAML-formatted strings.
    ndarray: numpy.ndarray
        Numpy array to be converted into a YAML-formatted string.

    Returns
    -------
    str
        YAML-formatted string representing this Numpy array.
    '''

    # Dictionary mapping from numpy.array() parameter names to values.
    array_params = {
        'object': ndarray.tolist(),
        'dtype': ndarray.dtype.name,
    }

    # Convert this dictionary into a stringified YAML mapping.
    return dumper.represent_mapping(_TAG_NUMPY_NDARRAY, array_params)


def _represent_numpy_bool(dumper, npbool: 'numpy.bool') -> str:
    '''
    Convert the passed Numpy boolean into a YAML-formatted string.

    Specifically, this function returns a stringified YAML boolean equivalent
    to the passed boolean.

    Parameters
    ----------
    dumper: yaml.Dumper
        Object converting arbitrary Python objects to YAML-formatted strings.
    npbool: numpy.bool
        Numpy boolean to be converted into a YAML-formatted string.

    Returns
    -------
    str
        YAML-formatted string representing this Numpy bool.
    '''

    return dumper.represent_bool(bool(npbool))


def _represent_numpy_complex(dumper, npcomplex: 'numpy.complex128') -> str:
    '''
    Convert the passed Numpy complex number into a YAML-formatted string.

    Specifically, this function returns a stringified YAML complex number
    equivalent to the passed complex number.

    Parameters
    ----------
    dumper: yaml.Dumper
        Object converting arbitrary Python objects to YAML-formatted strings.
    npcomplex: numpy.complex128
        Numpy complex number to be converted into a YAML-formatted string.

    Returns
    -------
    str
        YAML-formatted string representing this Numpy complex number.
    '''

    return dumper.represent_complex(complex(npcomplex))


def _represent_numpy_float(dumper, npfloat: 'numpy.float64') -> str:
    '''
    Convert the passed Numpy float into a YAML-formatted string.

    Specifically, this function returns a stringified YAML float equivalent to
    the passed float.

    Parameters
    ----------
    dumper: yaml.Dumper
        Object converting arbitrary Python objects to YAML-formatted strings.
    npfloat: numpy.float64
        Numpy float to be converted into a YAML-formatted string.

    Returns
    -------
    str
        YAML-formatted string representing this Numpy float.
    '''

    return dumper.represent_float(float(npfloat))


def _represent_numpy_int(dumper, npint: 'numpy.int64') -> str:
    '''
    Convert the passed Numpy integer into a YAML-formatted string.

    Specifically, this function returns a stringified YAML integer equivalent
    to the passed integer.

    Parameters
    ----------
    dumper: yaml.Dumper
        Object converting arbitrary Python objects to YAML-formatted strings.
    npint: numpy.int64
        Numpy integer to be converted into a YAML-formatted string.

    Returns
    -------
    str
        YAML-formatted string representing this Numpy integer.
    '''

    return dumper.represent_int(int(npint))



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

"""

Creates a mitochondria class, which includes a suite of mitochondria-specific molecules, along with
mitochondria-specific pumps (i.e. electron transport chain), channels, and specific methods.
This class also contains the facilities to initialize, define the core computations for a simulation loop,
remove mitochondria during a cutting event, save and report on data, and plot.

"""

import copy

import numpy as np

from betse.science import sim_toolbox as stb


class Mito(object):

    def __init__(self, sim, cells, p):

        # init basic fields
        self.mit_vol = 0.5*cells.cell_vol     # mit volume
        # print("!!!!!!!!! mit_vol: {}".format(self.mit_vol))
        self.mit_sa = 0.5*cells.cell_sa      # mit surface areas
        # self.Vmit = np.zeros(sim.cdl)   # initial transmembrane voltage for mit
        self.Vmit = np.zeros(sim.cdl)  # initial transmembrane voltage for mit
        self.Q = np.zeros(sim.cdl)     # total charge in mit
        self.cm_mit = self.mit_sa*p.cm    # mit membrane capacitance

        sim.cc_mit = copy.deepcopy(sim.cc_cells)    # ion concentrations
        self.Dm_mit = copy.deepcopy(sim.cc_cells)    # membrane permeability

        for arr in self.Dm_mit:

            arr[:] = 1.0e-19                 # membrane permeability altered so all are minimal

        # set calcium concentration in mitochondria to an initially low value:
        # if p.ions_dict['Ca'] == 1:
        #     self.Dm_mit[sim.iCa] = 1.0e-15  # add a mitochondrial calcium uniporter set
        #     # sim.cc_mit[sim.iCa] = 10.0e-6  # [100 nM]

        self.Dm_mit_base = copy.deepcopy(self.Dm_mit)  # copies of Dm for ion channel dynamics
        self.Dm_channels = copy.deepcopy(self.Dm_mit)

        self.zmit = copy.deepcopy(sim.cc_cells)

        for i, arr in enumerate(self.zmit):
            arr[:] = sim.zs[i]

    def get_v(self, sim, p):

        self.Q = stb.get_charge(sim.cc_mit, self.zmit, self.mit_vol, p) + self.extra_rho*self.mit_vol

        self.Vmit = (1/self.cm_mit)*self.Q

    def update(self, sim, cells, p):

        # self.channels(sim, cells, p)

        for i in sim.movingIons:

            IdCM = np.ones(sim.cdl)

            f_ED = stb.electroflux(sim.cc_cells[i], sim.cc_mit[i], self.Dm_mit[i], p.tm*IdCM, sim.zs[i]*IdCM,
                self.Vmit, sim.T, p, rho=1)

            # update with flux
            sim.cc_cells[i] = sim.cc_cells[i] - f_ED*(self.mit_sa/cells.cell_vol)*p.dt
            sim.cc_mit[i] = sim.cc_mit[i] + f_ED*(self.mit_sa/self.mit_vol)*p.dt

        self.get_v(sim, p)

    # def channels(self, sim, cells, p):

    # FIXME major useful channel would be the Ca2+ uniporter, which opens with 1 - 5 um cytosolic calcium
    #
    #     pass   # FIXME finish this

    def remove_mits(self, sim, target_inds_cell):

        # remove cells from the mit voltage list:
        vmit2 = np.delete(self.Vmit, target_inds_cell)
        # reassign the new data vector to the object:
        self.Vmit = vmit2

        mitv2 = np.delete(self.mit_vol, target_inds_cell)
        self.mit_vol = mitv2

        mitca2 = np.delete(self.mit_sa, target_inds_cell)
        self.mit_sa = mitca2

        Q2 = np.delete(self.Q, target_inds_cell)
        self.Q = Q2

        cm2 = np.delete(self.cm_mit, target_inds_cell)
        self.cm_mit = cm2

        cc_mit2 = []

        for i, arr in enumerate(sim.cc_mit):

            # remove cells from the mit ion array in sim:
            arr2 = np.delete(arr, target_inds_cell)
            cc_mit2.append(arr2)

        sim.cc_mit = np.asarray(cc_mit2)

        dmit1 = []
        dmit2 = []
        dmit3 = []
        zmit = []

        for i, arr in enumerate(self.Dm_mit):

            arr2 = np.delete(arr, target_inds_cell)
            dmit2.append(arr2)

        self.Dm_mit = np.asarray(dmit2)

        for i, arr in enumerate(self.Dm_mit_base):

            arr2 = np.delete(arr, target_inds_cell)
            dmit1.append(arr2)

        self.Dm_mit_base = np.asarray(dmit1)

        for i, arr in enumerate(self.Dm_channels):

            arr2 = np.delete(arr, target_inds_cell)
            dmit3.append(arr2)

        self.Dm_channels = np.asarray(dmit3)

        for i, arr in enumerate(self.zmit):

            arr2 = np.delete(arr, target_inds_cell)
            zmit.append(arr2)

        self.zmit = np.asarray(zmit)










#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **Numpy-specific scalar** (i.e., primitive non-standard data types
specific to the Numpy API) facilities.
'''

# ....................{ IMPORTS                           }....................
# import numpy as np
# from betse.util.io.log import logs
from betse.util.type.types import type_check, NumpyScalarType, ScalarTypes

# ....................{ CONVERTERS                        }....................
@type_check
def to_python(scalar: NumpyScalarType) -> ScalarTypes:
    '''
    Standard Numpy-agnostic scalar (e.g., :class:`bool`) losslessly coerced
    from the passed non-standard Numpy-specific scalar (e.g.,
    :class:`np.bool_`).

    This function converts the passed Numpy scalar to the corresponding Python
    scalar such that the latter is guaranteed to be a **lossless copy** (i.e.,
    copy with *no* appreciable loss in numerical precision) of the former.

    Parameters
    ----------
    scalar : NumpyScalarType
        Numpy-specific scalar to be converted into the corresponding
        Numpy-agnostic scalar.

    Returns
    ----------
    ScalarTypes
        Numpy-agnostic scalar converted from this Numpy-specific scalar.

    See Also
    ----------
    https://stackoverflow.com/a/11389998/2809027
        StackOverflow answer strongly inspiring this implementation.

    Examples
    ----------
        >>> from betse.lib.numpy import npscalar
        >>> import numpy as np
        >>> python_bool = True
        >>> numpy_bool = np.bool_(python_bool)
        >>> numpy_bool
        True
        >>> numpy_bool is True
        False
        >>> numpy_bool == True
        False
        >>> python_numpy_bool = npscalar.to_python(numpy_bool)
        >>> python_numpy_bool
        True
        >>> python_numpy_bool is True
        True
        >>> python_numpy_bool == True
        True
    '''

    # This is insanity. This is Numpy. For further details, see the official
    # documentation for the numpy.ndarray.item() method at:
    #     https://docs.scipy.org/doc/numpy/reference/generated/numpy.ndarray.item.html
    return scalar.item()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Project-wide **type hints** (i.e., PEP-compliant annotations of general-purpose
interest throughout the codebase).
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Annotated,
    ContextManager,
    Generator,
)
from beartype.vale import (
    IsAttr,
    IsEqual,
)
from numpy import ndarray

# ....................{ HINTS                              }....................
ContextManagerOrGenerator = ContextManager | Generator
'''
PEP-compliant type hint matching either a context manager *or* generator.
'''

# ....................{ HINTS ~ lib : numpy : 1d           }....................
NDArrayNdim1 = Annotated[ndarray, IsAttr['ndim', IsEqual[1]]]
'''
PEP-compliant type hint matching a 1-dimensional NumPy array.
'''


NDArrayNdim1Size2 = Annotated[
    ndarray,
    IsAttr['ndim', IsEqual[1]],
    IsAttr['size', IsEqual[2]],
]
'''
PEP-compliant type hint matching a 1-dimensional NumPy **2-array** (i.e., array
containing exactly two numbers, typically comprising the X and Y coordinates of
some 2-dimensional point).
'''

# ....................{ HINTS ~ lib : numpy : 2d           }....................
NDArrayNdim2 = Annotated[ndarray, IsAttr['ndim', IsEqual[2]]]
'''
PEP-compliant type hint matching a 2-dimensional NumPy array.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **test suite** (e.g., collection of one or more functional or unit
tests collectively exercising problematic features of this application)
facilities.
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.decorator.decmemo import func_cached

# ....................{ TESTERS                            }....................
@func_cached
def is_testing() -> bool:
    '''
    ``True`` only if the active Python interpreter is currently exercising
    tests (e.g., via the :mod:`pytest` test harness).

    Caveats
    ----------
    **This function is only safely callable after application initialization,**
    implying that this function should *never* be called from module scope.
    For generality, this function requires the application metadata singleton
    to have already been defined.

    Raises
    ----------
    BetseMetaAppException
        If this application is uninitialized (i.e., the application metadata
        singleton is currently undefined).
    '''

    # There are numerous means of implementing this test, including:
    #
    # * Testing whether or not the "pytest" package has been previously
    #   imported. This is the most commonly recommended approach on both
    #   StackOverflow and in Django commentary. While trivial to implement,
    #   this approach has the distinct disadvantage of generating spurious
    #   false positives. Avoiding such issues requires *NEVER* importing from
    #   the "pytest" package from the main codebase. To generalize testing
    #   functionality across this application and downstream consumers (e.g.,
    #   BETSEE), the "betse.util.test" subpackage commonly does just that.
    #   Ergo, this approach is effectively infeasible here.
    # * Enabling a private boolean constant (e.g.,
    #   "betse.util.test.tests._IS_TESTING") at "pytest" session start time
    #   (e.g., from a session-wide autouse fixture or
    #   betse_test.conftest.pytest_configure() hook). While technically
    #   feasible, this approach is considerably less trivial *AND* invites
    #   subtle complications at chicken-and-egg application initialization
    #   time. Quite simply: the rewards are low and the prices are high.
    #
    # The current approach circumvents all of the above issues. Like the first
    # approach above, this approach is trivial to implement; unlike the first
    # approach above, this approach does *NOT* generate spurious false
    # positives. Why? Because the main codebase is guaranteed to *NEVER* import
    # from this application's test suite. Doing so would fundamentally violate
    # sanity in numerous ways and, in any case, is never desirable. In
    # particular, this application's test suite is *NOT* bundled with this
    # application and hence *ONLY* importable during development.

    # Avoid circular import dependencies.
    from betse.util.app.meta import appmetaone
    from betse.util.py.module import pymodname

    # Name of the root package of this application's test suite.
    test_package_name = appmetaone.get_app_meta().test_package_name

    # Return true only if that package has been previously imported from.
    return pymodname.is_imported(test_package_name)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level logging filter subclasses.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific modules at the top-level -- excluding those explicitly
# known *NOT* to import from this module. Since all application-specific modules
# must *ALWAYS* be able to safely import from this module at any level, these
# circularities are best avoided here rather than elsewhere.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from betse import metadata
from betse.util.io.log.logenum import LogLevel
from betse.util.type.types import type_check
from logging import Filter, LogRecord

# ....................{ CLASSES                            }....................
class LogFilterThirdPartyDebug(Filter):
    '''
    Log filter ignoring all log records with logging levels less than or equal
    to :attr:`LogLevel.DEBUG` *and* names not prefixed by ``betse``.

    Equivalently, this log filter *only* retains log records with either:

    * Logging levels greater than :attr:`LogLevel.DEBUG`.
    * Names prefixed by ``betse``, including both:
      * ``betse``, the top-level package for BETSE.
      * ``betsee``, the top-level package for BETSEE.

    This log filter prevents ignorable debug messages logged by third-party
    frameworks (e.g., Pillow) from polluting this application's debug output.
    '''

    @type_check
    def filter(self, log_record: LogRecord) -> bool:
        '''
        ``True`` only if the passed log record is to be retained.
        '''

        # print('log record name: {}'.format(log_record.name))
        return (
            log_record.levelno > LogLevel.DEBUG or
            log_record.name.startswith(metadata.PACKAGE_NAME))


class LogFilterMoreThanInfo(Filter):
    '''
    Log filter ignoring all log records with logging levels greater than
    :attr:`LogLevel.INFO``.

    Equivalently, this log filter *only* retains log records with logging levels
    less than or equal to :attr:`LogLevel.INFO``.
    '''

    @type_check
    def filter(self, log_record: LogRecord) -> bool:
        '''
        ``True`` only if the passed log record is to be retained.
        '''

        return log_record.levelno <= LogLevel.INFO



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Controls a gene regulatory network.

Creates and electrodiffuses a suite of customizable general gene products in
the BETSE ecosystem, where the gene products are assumed to activate and/or
inhibit the expression of other genes (and therefore the production of other
gene products) in the gene regulatory network (GRN).
'''

#FIXME: Unify the large amount of code shared in common between this and the
#"gene" submodule. See the latter for details.

# ....................{ IMPORTS                           }....................
import numpy as np
from betse.science import filehandling as fh
from betse.science.chemistry.networks import MasterOfNetworks
from betse.science.chemistry.netplot import set_net_opts
from betse.science.phase.phasecls import SimPhase
from betse.util.io.log import logs
from betse.util.path import pathnames
from betse.util.type.types import type_check

# ....................{ CLASSES                           }....................
class MasterOfMolecules(object):

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, p):

        #FIXME: Extract the "GeneralNetwork.betse" basename into a new
        #configuration option, perhaps in the "init file saving" section.
        #FIXME: Replace "GeneralNetwork.betse" with "GeneralNetwork.betse.gz" to
        #compress this pickled file.

        # Define data paths for saving an initialization and simulation run:
        self.savedMoM = pathnames.join(p.init_pickle_dirname, 'GeneralNetwork.betse')


    #FIXME: Most of this method appears to have been copy-and-pasted
    #from the read_mol_config() method below. That's... not the best. Let's
    #extract the code shared in common between these two methods into a new
    #_init_mols() method internally called by these two methods.
    @type_check
    def reinitialize(self, phase: SimPhase) -> None:

        # Localize high-level phase objects for convenience.
        cells = phase.cells
        p     = phase.p
        sim   = phase.sim

        # read the config file into a dictionary:
        config_dic = p.network_config

        # Time dilation:
        self.core.time_dila = float(config_dic.get('time dilation factor', 1.0))

        # obtain specific sub-dictionaries from the config file:
        substances_config = config_dic['biomolecules']
        reactions_config = config_dic.get('reactions', None)
        transporters_config = config_dic.get('transporters', None)
        channels_config = config_dic.get('channels', None)
        modulators_config = config_dic.get('modulators', None)

        self.core.tissue_init(phase, substances_config)

        if reactions_config is not None:
            # initialize the reactions of metabolism:
            self.core.read_reactions(reactions_config, sim, cells, p)
            self.core.write_reactions()
            self.core.create_reaction_matrix()
            self.core.write_reactions_env()
            self.core.create_reaction_matrix_env()

            self.reactions = True

        else:
            self.core.create_reaction_matrix()
            self.core.create_reaction_matrix_env()
            self.reactions = False

        # initialize transporters, if defined:
        if transporters_config is not None:
            self.core.read_transporters(transporters_config, phase)
            self.core.write_transporters(sim, cells, p)

            self.transporters = True

        else:
            self.transporters = False

        # initialize channels, if desired:
        if channels_config is not None:
            self.core.read_channels(channels_config, phase)
            self.channels = True

        else:
            self.channels = False

        # initialize modulators, if desired:
        if modulators_config is not None:
            self.core.read_modulators(modulators_config, sim, cells, p)
            self.modulators = True

        else:
            self.modulators = False

    # ..................{ READERS                           }..................
    @type_check
    def read_mol_config(self, phase: SimPhase) -> None:

        # Localize high-level phase objects for convenience.
        cells = phase.cells
        p     = phase.p
        sim   = phase.sim

        # read the config file into a dictionary:
        config_dic = p.network_config

        # obtain specific sub-dictionaries from the config file:
        substances_config = config_dic['biomolecules']
        reactions_config = config_dic.get('reactions', None)
        transporters_config = config_dic.get('transporters', None)
        channels_config = config_dic.get('channels', None)
        modulators_config = config_dic.get('modulators', None)

        # initialize the substances of metabolism in a core field encapsulating
        # Master of Molecules:
        self.core = MasterOfNetworks(sim, cells, substances_config, p)

        # Time dilation:
        self.core.time_dila = float(config_dic.get('time dilation factor', 1.0))

        # read in substance properties from the config file, and initialize basic properties:
        self.core.read_substances(sim, cells, substances_config, p)
        self.core.tissue_init(phase, substances_config)

        if reactions_config is not None:
            # initialize the reactions of metabolism:
            self.core.read_reactions(reactions_config, sim, cells, p)
            self.core.write_reactions()
            self.core.create_reaction_matrix()
            self.core.write_reactions_env()
            self.core.create_reaction_matrix_env()

            self.reactions = True
        else:
            self.core.create_reaction_matrix()
            self.core.create_reaction_matrix_env()
            self.reactions = False

        # initialize transporters, if defined:
        if transporters_config is not None:
            self.core.read_transporters(transporters_config, phase)
            self.core.write_transporters(sim, cells, p)

            self.transporters = True
        else:
            self.transporters = False

        # initialize channels, if desired:
        if channels_config is not None:
            self.core.read_channels(channels_config, phase)
            self.channels = True
        else:
            self.channels = False

        # initialize modulators, if desired:
        if modulators_config is not None:
            self.core.read_modulators(modulators_config, sim, cells, p)
            self.modulators = True
        else:
            self.modulators = False

        # read in network plotting options:
        self.core.net_plot_opts = config_dic.get('network plotting', None)

        # set plotting options for the network:
        set_net_opts(self.core, self.core.net_plot_opts, p)

        # after primary initialization, check and see if optimization required:
        optim_exists = config_dic.get('optimization', None)
        if optim_exists is not None:
            opti = config_dic['optimization']['optimize network']
            self.core.opti_N = config_dic['optimization']['optimization steps']
            self.core.opti_method = config_dic['optimization']['optimization method']
            self.core.target_vmem = float(config_dic['optimization']['target Vmem'])
            self.core.opti_T = float(config_dic['optimization']['optimization T'])
            self.core.opti_step = float(config_dic['optimization']['optimization step'])

            if opti:
                logs.log_info(
                    'Analyzing the general network for optimal rates...')
                self.core.optimizer(sim, cells, p)
                self.reinitialize(phase)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level simulation phase classes.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseSimPhaseException
from betse.science.phase import phasecallbacks
from betse.science.phase.phasecallbacks import SimCallbacksBCOrNoneTypes
from betse.science.enum.enumphase import SimPhaseKind
# from betse.util.type.iterable import iterables
from betse.util.type.text.string import strjoin
from betse.util.type.types import type_check, NoneType

# ....................{ CLASSES                           }....................
#FIXME: Generalize to support dynamically changing cell structure as follows:
#
#* Define a new "cells_time" list indexed by time step for the current phase,
#  each item of which is a "Cells" instance. Most such items are simply
#  references to "cells"; all other items if any will be references to a
#  post-cutting event "Cells" instance. Edge cases include:
#  * For the seed phase, "cells_time" should be "None".
#  * For the init phase, "cells_time" should be a list of the expected length
#    (i.e., the number of initialization time steps), each of whose items is
#    an unconditional reference to "cells".
#  * For the sim phase, "cells_time" should be a list of the expected length
#    (i.e., the number of simulation time steps), only the first of whose items
#    is guaranteed to be a reference to "cells".
#* Sadly, most of the existing "cache" object needs to be refactored away. We
#  currently cache a large number of cell data structures under the assumption
#  that the cell structure remains constant. Clearly, this is *NOT* the case;
#  this was a bad assumption. Ergo, we need to dynamically recompute at each
#  time step most of the data that we currently cache in the "cache" object on
#  the first time step. (History is a collection of tragic mistakes.)
#FIXME: While effectively mandatory, the above generalization opens up the
#proverbial can of worms. In particular: pickling. We currently only pickle
#"cells", "p", and "sim". That's an obvious problem. Fortunately, the
#answer is simple (albeit arguably non-ideal): just define a new
#"Simulator.cells_time" list in lieu of a "SimPhase.cells_time" list, but
#otherwise defined exactly as above.

class SimPhase(object):
    '''
    High-level simulation phase, encapsulating all lower-level objects required
    to perform a single phase (e.g., seed, initialization, simulation) of a
    given cell cluster, configuration, and simulation.

    This object principally behaves as a simple container whose:

    * Direct parent is the root-level
      :class:`betse.science.simrunner.SimRunner` object owning all objects
      pertaining to the current cell cluster, configuration, and simulation.
    * Direct children are:

      * The current cell cluster.
      * The current simulation.
      * The current simulation configuration.

    Attributes
    ----------
    kind : SimPhaseKind
        Type of this simulation phase.

    Attributes (High-level)
    ----------
    cells : betse.science.cells.Cells
        Cell cluster for this phase. If this phase is currently being
        simulated, this object refers to the cluster at the time step being
        simulated; else (i.e., if this phase has already been simulated), this
        object refers to the cluster at the last simulation time step for this
        phase. Since the cluster may change with time while simulating (e.g.,
        due to surgical interventions like cutting events), this object applies
        *only* to this time step. Callers should *not* assume this object to
        uniformly apply to any other time steps of either this or other phases;
        callers should index the :attr:`sim.cells_time` list by time step for
        the cluster at the time step immediately following that time step.
    dyna : betse.science.tissue.tishandler.TissueHandler
        Tissue handler for this phase.
    p : betse.science.parameters.Parameters
        Simulation configuration for this phase.
    sim : betse.science.sim.Simulator
        Simulation for this phase.
    cache : betse.science.phase.cache.cacheabc.SimPhaseCaches
        Simulation cache for this phase.

    Attributes (Low-level: Caller)
    ----------
    callbacks : SimCallbacksBC
        Caller-defined object whose methods are periodically called during this
        phase (e.g., to notify this caller of phase progress).

    Attributes (Low-level: Path)
    ----------
    export_dirname : StrOrNoneTypes
        Absolute path of the top-level directory containing all exported
        results (e.g., plots, animations, CSVs) for this simulation phase if
        this phase is either an initialization or simulation *or* ``None``
        otherwise (i.e., if this phase is a seed).
    '''

    # ..................{ INITIALIZORS                      }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        kind: SimPhaseKind,
        p: 'betse.science.parameters.Parameters',

        # Optional parameters.
        cells: ('betse.science.cells.Cells', NoneType) = None,
        sim:   ('betse.science.sim.Simulator', NoneType) = None,
        callbacks: SimCallbacksBCOrNoneTypes = None,
    ) -> None:
        '''
        Initialize this simulation phase.

        Parameters
        ----------
        kind : SimPhaseKind
            Current simulation phase type.
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        cells : (betse.science.cells.Cells, NoneType)
            Current cell cluster. Defaults to ``None``, in which case this
            defaults to the empty cell cluster for this configuration.
        sim : (betse.science.sim.Simulation, NoneType)
            Current simulation. Defaults to ``None``, in which case this
            defaults to an uninitialized simulation for this configuration.
        callbacks : SimCallbacksBCOrNoneTypes
            Caller-defined object whose methods are periodically called during
            this phase (e.g., to notify this caller of phase progress).
            Defaults to ``None``, in which case this defaults to a placeholder
            object whose methods all silently reduce to noops.
        '''

        # Avoid circular import dependencies.
        from betse.science.cells import Cells
        from betse.science.math.cache.cacheabc import SimPhaseCaches
        from betse.science.sim import Simulator
        from betse.science.tissue.tishandler import TissueHandler

        # Default all unpassed parameters to sane defaults.
        if callbacks is None:
            callbacks = phasecallbacks.make_default()
        if cells is None:
            cells = Cells(p=p)
        if sim is None:
            sim = Simulator()

        # Classify all passed parameters.
        self.callbacks = callbacks
        self.cells = cells
        self.kind = kind
        self.p = p
        self.sim = sim

        #FIXME: To ensure the "dyna" object is fully initialized, refactor the
        #"TissueHandler" class as follows:
        #
        #* The TissueHandler.init_profiles() method should be refactored into a
        #  private TissueHandler._map_tissues_to_cells() method having
        #  the following signature:
        #      @type_check
        #      def _map_tissues_to_cells(self, phase: SimPhase) -> None:
        #* The TissueHandler.__init__() method should be refactered to:
        #  * As the very first logic in that method, validate that the passed
        #    "cells" and "sim" objects have been *FULLY* initialized.
        #  * As the very last logic in that method, call the newly refactored
        #    self._map_tissues_to_cells() method.
        #* The TissueHandler._map_tissues_to_cells() method should *NOT* be
        #  called by any other class. In practice, we probably will need to
        #  break encapsulation and call the newly private
        #  TissueHandler._map_tissues_to_cells() method elsewhere. When we do
        #  so, add a FIXME comment suggesting this to be bad.
        #
        #Alternately, if the _map_tissues_to_cells() method truly *DOES* need
        #to be called elsewhere, simply make it public. *sigh*

        # Classify all remaining high-level objects for this phase.
        self.cache = SimPhaseCaches(phase=self)
        self.dyna = TissueHandler(p=p)

        # Initialize all kludges required by this phase.
        self._init_kludge()

        #FIXME: Isolate exports produced by the "seed" phase to their own
        #directory; for simplicity, these exports currently reuse the same
        #directory as that of the "init" phase.

        # Absolute path of the top-level exports directory for this phase.
        if kind is SimPhaseKind.SEED:
            self.export_dirname = p.init_export_dirname
        elif kind is SimPhaseKind.INIT:
            self.export_dirname = p.init_export_dirname
        elif kind is SimPhaseKind.SIM:
            self.export_dirname = p.sim_export_dirname
        else:
            raise BetseSimPhaseException(
                'Simulation phase "{}" unrecognized.'.format(kind.name))

    # ..................{ INITIALIZORS ~ kludge             }..................
    #FIXME: Refactor away all kludges implemented by this method as detailed.
    def _init_kludge(self) -> None:
        '''
        Initialize all **kludges** (i.e., inelegant short-term workarounds
        intended to be replaced by actual long-term solutions) required by this
        simulation phase.
        '''

        #FIXME: Eliminate this crude hack by refactoring all references to
        #"sim.dyna" throughout the codebase to "phase.dyna" instead. Naturally,
        #this will require refactoring all methods referencing "sim.dyna" to
        #accept a "phase: SimPhase" parameter. After doing so, remove this line.
        #Praise be to the multifoliate rose!
        #FIXME: O.K.; thanks to intrepid obsessiveness, only one submodule
        #referencing "sim.dyna" exists in the codebase: the
        #"betse.science.chemistry.networks" submodule. There currently exist
        #only 6 references to "sim.dyna" in this submodule. (Make 'em vanish.)
        self.sim.dyna = self.dyna

        #FIXME: Eliminate this crude hack by refactoring all references to
        #"p._run_sim" throughout the codebase to test "phase.kind" instead.
        #Although there only exists one remaining reference, refactoring that
        #reference will be a living nightmare. Why? Because the reference
        #resides in a modulator function -- which are currently called in an
        #extremely dynamic, inscrutable, non-greppable manner. Refactoring this
        #single reference to test "phase.kind" instead will require:
        #
        #* Defining a new high-level
        #  betse.science.math.modulate.make_modulator() function resembling:
        #      def make_modulator(modulator_name: str) -> CallableTypes
        #* Refactoring *ALL* modulator functions to accept a single "phase:
        #  SimPhase" parameter rather than the two "cells, p" parameters
        #  currently accepted by modulator functions.
        #* Refactoring *ALL* calls to modulator functions to instead:
        #  * First call the make_modulator() function. This will ensure that
        #    this *NEVER* happens again, by ensuring that we're able to
        #    reliably manage modulator functions from a single point of
        #    failure.
        #  * Next call the callable returned by that function, passing that
        #    callable the current "phase" object. Of course, this will in turn
        #    necessitate refactoring each callable performing such a call to
        #    transitively accept the current "phase" object. This is Hell.
        #* Refactoring the f_sweep() modulator function to test "phase.kind"
        #  rather than "p._run_sim".
        #
        #To grep for existing calls to modulator functions, grep for both the
        #strings "\bmodulator function\b" and "\bgetattr\(". Our YAML structure
        #appears to reliably refer to the desired modulator function for a
        #given operation via "'modulator function'". *sigh*
        self.p._run_sim = self.kind is SimPhaseKind.SIM

        # Initialize all time-specific kludges required by this phase.
        self._init_kludge_time()


    def _init_kludge_time(self) -> None:
        '''
        Initialize all time-specific kludges required by this simulation phase.
        '''

        # If this is the seed phase, avoid defining any of the following
        # ad-hoc "Parameters" instance variables.
        if self.kind is SimPhaseKind.SEED:
            return

        #FIXME: Eliminate all of the following crude hacks by:
        #
        #* Defining a new "betse.science.phase.phasetime" submodule.
        #* Defining a new "SimPhaseTime" class in this submodule: e.g.,
        #    class SimPhaseTime(object):
        #        '''
        #        Attributes
        #        ----------
        #        dt : float
        #            Duration in seconds of each time step (including sampled and unsampled)
        #            for the current simulation phase.
        #        t_resample : float
        #            Number of time steps between each sampled time step, including that
        #            sampled time step itself. Notably, if the current time step ``t`` is a
        #            sampled time step:
        #
        #            * ``t + t_resample`` is the next sampled time step (if any).
        #            * ``t - t_resample`` is the prior sampled time step (if any).
        #        total_time : float
        #            Duration in seconds of the current simulation phase *not* modified by
        #            extraneous settings (e.g., gap junction acceleration factor).
        #        '''
        #
        #* Defining a new "time" variable of this "SimPhase" instance in the
        #  above __init__() method: e.g.,
        #      from betse.science.phase.phasetime import SimPhaseTime
        #      self.time = SimPhaseTime()
        #* Shifting all of the following instance variables from being
        #  dynamically defined on the "Parameters" instance here to being
        #  statically (i.e., normally) defined in the SimPhaseTime.__init__()
        #  method instead.
        #
        #Lastly, note that if this is the seed phase, that the "self.time"
        #variable should default to "None" to induce exceptions on any
        #erroneous attempt to access time parameters when seeding.

        # Set temporal attributes specific to the passed simulation phase type.
        # These attributes include:
        #
        # * "dt", the duration in seconds of each time step for this phase.
        # * "total_time", the duration in seconds of this phase.

        # If this phase is an initialization...
        if self.kind is SimPhaseKind.INIT:
            self.p.dt = self.p.init_time_step
            self.p.resample = self.p.init_time_sampling
            self.p.total_time = self.p.init_time_total
        # Else if this phase is a simulation...
        elif self.kind is SimPhaseKind.SIM:
            self.p.dt = self.p.sim_time_step
            self.p.resample = self.p.sim_time_sampling
            self.p.total_time = self.p.sim_time_total
        else:
            raise BetseSimPhaseException(
                'Simulation phase "{}" unsupported.'.format(self.kind.name))

        # Number of time steps (including sampled and unsampled) between each
        # unsampled time step, including that unsampled time step itself.
        self.p.t_resample = self.p.resample / self.p.dt

    # ..................{ EXCEPTIONS                        }..................
    @type_check
    def die_unless_kind_seed(self) -> None:
        '''
        Raise an exception unless this is the seed phase.

        Raises
        ----------
        BetseSimPhaseException
            If this phase is *not* a seed.
        '''

        self._die_unless_kind(SimPhaseKind.SEED)


    @type_check
    def die_unless_kind_init_or_sim(self) -> None:
        '''
        Raise an exception unless this is either the initialization *or*
        simulation phases.

        Raises
        ----------
        BetseSimPhaseException
            If this phase is neither an initialization *or* simulation.
        '''

        self._die_unless_kind(SimPhaseKind.INIT, SimPhaseKind.SIM)

    # ..................{ EXCEPTIONS ~ private              }..................
    @type_check
    def _die_unless_kind(self, *kinds: SimPhaseKind) -> None:
        '''
        Raise an exception unless the type of this simulation phase is in the
        passed tuple of simulation phase types.

        Parameters
        ----------
        kinds : tuple[SimPhaseKind]
            Tuple of all simulation phase types (i.e., members of the
            :class:`SimPhaseKind` enumeration) to test this simulation phase
            against.

        Raises
        ----------
        BetseSimPhaseException
            If this simulation phase's type is *not* in the passed tuple.
        '''

        # If the type of this simulation phase is *NOT* a passed type, raise an
        # exception.
        if self.kind not in kinds:
            # Generator comprehension yielding the machine-readable name of
            # each of the passed types.
            kinds_name = (str(kind) for kind in kinds)

            # Human-readable double-quoted disjunction of these names.
            kinds_name_joined = strjoin.join_as_disconjunction_double_quoted(
                *kinds_name)

            # Raise an exception embedding this disjunction.
            raise BetseSimPhaseException(
                'Simulation phase "{}" not {}.'.format(
                    self.kind, kinds_name_joined))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level module and package name-based facilities.

All functions defined by this submodule require that modules be passed as
fully-qualified names (i.e., ``.``-delimited).

See Also
--------
:mod:`betse.util.py.module.pymodule`
    Related submodule whose functions accept already imported module objects.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on missing mandatory dependencies,
# the top-level of this module may import *ONLY* from packages guaranteed to
# exist at installation time -- which typically means *ONLY* BETSE packages and
# stock Python packages.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import importlib, sys
from beartype.typing import (
    Collection,
    Dict,
)
from betse.exceptions import BetseModuleException
from betse.util.io.log import logs
from betse.util.type.iterable.mapping.mapcls import DefaultDict
from betse.util.type.types import (
    type_check, MappingOrNoneTypes, ModuleType, StrOrNoneTypes)
from collections import (
    defaultdict,
    namedtuple,
)
from importlib import util as importlib_util
from importlib.machinery import ModuleSpec

# ....................{ CLASSES                            }....................
DependencyCommand = namedtuple('DependencyCommand', ('name', 'basename',))
DependencyCommand.__doc__ = '''
    Lightweight metadata describing a single external command required by an
    application dependency of arbitrary type (including optional, mandatory,
    runtime, testing, and otherwise).

    Attributes
    ----------
    name : str
        Human-readable name associated with this command (e.g., ``Graphviz``).
    basename : str
        Basename of this command to be searched for in the current ``${PATH}``.
    '''

# ....................{ GLOBALS                            }....................
DEPENDENCY_TO_COMMANDS: Dict[str, Collection[DependencyCommand]] = {
    # "pydot" requires Graphviz (i.e., "dot") to be externally installed.
    'pydot': (DependencyCommand(name='Graphviz', basename='dot'),),
}
'''
Dictionary mapping each **runtime dependency name** (i.e., human-readable name
of an open-source project hosted by the Python Packaging Index (PyPI), such as
``"PyYAML"``, of a runtime dependency of this project) to a collection of one
or more external commands required by that project.

See Also
--------
:download:`/doc/md/INSTALL.md`
    Human-readable list of these dependencies.
'''


DEPENDENCY_TO_MODULE_NAME = DefaultDict(
    missing_key_value=lambda self, missing_key: missing_key,
    initial_mapping={
        'Numpy':  'numpy',
        'Pillow': 'PIL',
        'pillow': 'PIL',
        'PyYAML': 'yaml',
        'pyyaml': 'yaml',
        'SciPy':  'scipy',
    }
)
'''
Dictionary mapping each **runtime dependency name** (i.e., human-readable name
of an open-source project hosted by the Python Packaging Index (PyPI), such as
``"PyYAML"``, of a runtime dependency of this project) to the fully-qualified
name of the module or package providing that project (e.g., ``yaml``).

All modules and packages explicitly unmapped by this dictionary default to the
identity mapping -- that is, mapping the fully-qualified names of those modules
and packages to themselves (e.g., from ``pytest`` to ``pytest``).
'''


MODULE_NAME_TO_VERSION_ATTR_NAME = defaultdict(
    # Default attribute name to be returned for all unmapped modules.
    lambda: '__version__',

    #FIXME: Sample entry preserved purely for posterity.
    # Module-specific attribute names.
    # PIL='PILLOW_VERSION',
)
'''
Dictionary mapping the fully-qualified name of a module or package (e.g.,
:mod:``PIL``) to the name of the attribute (e.g., ``PILLOW_VERSION``) declared
by that module or package, providing that module or package's version
specifier.

All modules and packages explicitly unmapped by this dictionary default to the
canonical ``__version__`` attribute name.
'''

# ....................{ EXCEPTIONS                         }....................
@type_check
def die_if_module(module_name: str) -> None:
    '''
    Raise an exception if the module with the passed name **exists** (i.e., is
    importable by the active Python interpreter).

    Parameters
    ----------
    module_name : str
        Fully-qualified name of the module to be validated.

    Raises
    ------
    BetseModuleException
        If this module exists.

    See Also
    --------
    :func:`.is_module`
        Further details.
    '''

    # If this module is importable, raise an exception.
    if is_module(module_name):
        raise BetseModuleException(
            'Module "{}" already exists.'.format(module_name))


@type_check
def die_unless_module(
    module_name: str, exception_message: StrOrNoneTypes = None) -> None:
    '''
    Raise an exception with the passed message (defaulting to a message
    synthesized from the passed module name) if the module with the passed name
    does *not* **exist** (i.e., is unimportable by the active Python
    interpreter).

    Parameters
    ----------
    module_name : str
        Fully-qualified name of the module to be validated.
    exception_message : StrOrNoneTypes
        Message of the exception to be raised if this module does *not* exist.
        Defaults to ``None``, in which case a message is synthesized from this
        module name.

    Raises
    ------
    ImportError
        If this module either does not exist *or* does exist but is
        unimportable (e.g., due to module-scoped side effects at importation
        time). To allow callers to transparently handle importation errors in
        the standard way, this function raises a standard exception rather than
        an application-specific exception (e.g.,
        :class:`betse.exceptions.BetseModuleException`).

    See Also
    --------
    :func:`.is_module`
        Further details.
    '''

    # If this module is unimportable, raise an exception.
    if not is_module(module_name):
        # If no exception message was passed, synthesize one from this name.
        if not exception_message:
            exception_message = 'Module "{}" not found.'.format(module_name)

        # Raise this exception. (See the docstring for further details.)
        raise ImportError(exception_message)

# ....................{ TESTERS                            }....................
@type_check
def is_module(module_name: str) -> bool:
    '''
    ``True`` only if the module with the passed fully-qualified name **exists**
    (i.e., is importable under the active Python interpreter).

    Caveats
    -------
    In common edge cases, **this function may import all parent modules of this
    module as well as this module itself as a side effect.** Specifically, if
    this module is:

    * A **submodule** (i.e., contains a ``.`` character), this function
      necessarily imports *all* parent modules of this module.
    * *Not* importable via standard mechanisms (e.g., the OS X-specific
      :mod:`PyObjCTools` package), this function may import this module itself.

    Parameters
    ----------
    module_name : str
        Fully-qualified name of the module to be tested.

    Returns
    -------
    bool
        ``True`` only if this module exists.
    '''

    # Depending on context, this function behaves in one of three distinct
    # ways:
    #
    # * If this module's name is a key in the canonical dictionary
    #   "sys.modules" and has thus already been imported at least once under
    #   the active Python process, then...
    #   * If the "sys.modules[module_name].__spec__" attribute is set to a
    #     non-None value, that value is returned.
    #   * Else, the "ValueError" exception is raised.
    # * Else if this module is loadable by iteratively querying all module
    #   loaders in "sys.meta_path" (the canonical list of such loaders), a new
    #   spec is created describing this module and returned.
    # * Else this module is unloadable. In this case:
    #   * If this module is a submodule (i.e., this module's name contains a
    #     "." delimiter) and any parent module of this submodule is unloadable,
    #     the "ImportError" exception is raised.
    #   * Else, None is returned.
    #
    # Since this function only returns a boolean value, the above non-boolean
    # values and exceptions are converted into a simple boolean.
    try:
        return importlib_util.find_spec(module_name) is not None
    # If this module is a submodule (i.e., this module's name contains a "."
    # delimiter) and any parent module of this submodule is unloadable, this
    # submodule itself is unloadable.
    except ImportError:
        return False
    # If this module appears to have been imported at least once under the
    # active Python process but has no "__spec__" attribute, inspect deeper.
    # This exception does *NOT* necessarily imply this module to not exist.
    # This module may exist even if this exception is raised (e.g., for modules
    # dynamically defined at runtime rather than loaded from external files).
    #
    # Unfortunately, this exception does imply that conventional alternatives
    # to the prior function call (e.g., testing tuples generated by
    # pkgutil.iter_modules()) will also fail to find this module. As a
    # fallback, attempt to manually import this module. Since doing so
    # implicitly imports the "__init__.py" files of all parent packages of this
    # module and hence may have unhelpful side effects, we do so only if the
    # prior call failed.
    except ValueError:
        try:
            importlib.import_module(module_name)
            return True
        except ImportError:
            return False


@type_check
def is_imported(*module_names: str) -> bool:
    '''
    ``True`` only if all modules with the passed fully-qualified names have
    already been imported in the active Python process.
    '''

    # all(). It is awesome.
    return all(module_name in sys.modules for module_name in module_names)

# ....................{ GETTERS                            }....................
@type_check
def get_parent_module_name_or_none(module_name: str) -> StrOrNoneTypes:
    '''
    Name of the parent module of the child module with the passed name if this
    module has a parent *or* ``None`` otherwise (i.e., if this is a top-level
    module).

    Parameters
    ----------
    module_name : str
        Fully-qualified name of the module to be munged.

    Returns
    -------
    StrOrNoneTypes
        Either:

        * If this is a top-level module, ``None``.
        * Else, the fully-qualified name of the parent module of this module.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # Return the prefix of this module name preceding the last "." delimiter.
    return strs.get_prefix_or_none(
        text=module_name, anchor='.', is_first=False)

# ....................{ IMPORTERS                          }....................
@type_check
def import_module(
    module_name: str, exception_message: StrOrNoneTypes = None) -> ModuleType:
    '''
    Dynamically import and return the module, package, or C extension with the
    passed fully-qualified name if importable *or* raise an exception with the
    passed message otherwise.

    Parameters
    ----------
    module_name : str
        Fully-qualified name of the module to be validated.
    exception_message : StrOrNoneTypes
        Message of the exception to be raised if this module does *not* exist.
        Defaults to ``None``, in which case a message is synthesized from this
        module name.

    Raises
    ------
    ImportError
        If this module either does not exist *or* does exist but is
        unimportable (e.g., due to module-scoped side effects at importation
        time).
    '''

    #FIXME: This is extraneous and should thus be removed. The
    #importlib.import_module() function called below explicitly raises a
    #well-named "ModuleNotFoundError" exception if this module is unimportable.
    #Indeed, the is_module() function called by the die_unless_module()
    #function is suspiciously complex and probably broken in various edge
    #cases, particularly under newer Python interpreters.

    # If this module is unimportable, raise an exception.
    die_unless_module(module_name, exception_message)

    # Else, import and return this module.
    return importlib.import_module(module_name)


@type_check
def unimport_module_if_imported(*module_names: str) -> None:
    '''
    Dynamically unimport each of the modules, packages, or C extensions with
    the passed fully-qualified names that have been previously imported under
    the active Python interpreter.

    Specifically, for each passed name:

    * If this name is that of a previously imported module, this module is
      removed from the canonical :attr:`sys.modules` cache of all previously
      imported modules. On the next attempt to import this module, Python will
      re-import and re-cache that module to this cache.
    * Else, this module is silently ignored.
    '''

    # For each of the passed module names...
    for module_name in module_names:
        # If this module has already been imported, unimport this module.
        if is_imported(module_name):
            logs.log_debug('Unimporting module "{}"...'.format(module_name))
            del sys.modules[module_name]
        # Else, this module has *NOT* yet been imported. Ignore this module.

# ....................{ MAKERS                             }....................
@type_check
def make_module(
    module_name: str,
    module_attr_name_to_value: MappingOrNoneTypes = None,
    module_doc: StrOrNoneTypes = None,
    is_importable: bool = False,
) -> ModuleType:
    '''
    Dynamically create and return a new module with the passed name and
    optional passed docstring and attributes, optionally added to the standard
    :attr:`sys.modules` dictionary for external importation elsewhere.

    Caveats
    ----------
    **This function currently creates only non-package modules.** Although this
    function could be probably generalized to support dynamic creation of
    packages as well, packages impose certain real-world constraints *not*
    imposed by non-package modules (e.g., loading package submodules).

    Parameters
    ----------
    module_name : str
        Fully-qualified name of the module to be created.
    module_attr_name_to_value : MappingOrNoneTypes
        Dictionary mapping from the name to value of each module-scoped
        attribute to be declared in this module. Defaults to ``None``, in which
        case this module is empty (i.e., contains *no* attributes) by default.
    module_doc: StrOrNoneTypes
        **Docstring** (i.e., human-readable documentation in reStructuredText
        (reST) format) to be associated with this module. Defaults to ``None``,
        in which case this module is undocumented by default.
    is_importable : bool
        ``True`` only if external callers are allowed to trivially import this
        module via this module name elsewhere in this codebase. Note that, in
        this case, *all* parent packages of this module must already exist.
        Defaults to ``False`` for safety, in which case the returned object is
        the *only* initial reference to this module.

    Raises
    ------
    BetseModuleException
        If a module with this name already exists.
    ImportError
        If the ``is_importable`` parameter is ``True`` *and* one or more parent
        packages of this module do *not* already exist.

    See Also
    --------
    https://stackoverflow.com/questions/2931950/dynamic-module-creation
        StackOverflow *question* strongly inspiring this implementation. Note
        that, against all expectations, this question is substantially more
        informative than its answers.
    '''

    # If this module already exists, raise an exception.
    die_if_module(module_name)

    # Name of the parent module of this module if any *OR* None otherwise.
    parent_module_name = get_parent_module_name_or_none(module_name)

    # If this module is to be importable *AND* has a parent, raise an exception
    # if this parent does *NOT* already exist. By virtue of import mechanics,
    # this suffices to also recursively validate that *ALL* transitive parents
    # of this parent are also importable.
    if is_importable and parent_module_name is not None:
        die_unless_module(parent_module_name)

    # Module specification (i.e., metadata object describing this module).
    #
    # Note that module specifications are typically created and consumed by
    # low-level import machinery. In this case, however, we are creating a
    # module specification for the sole purpose of directly passing that same
    # module specification to the importlib.util.module_from_spec() function.
    module_spec = ModuleSpec(
        # Fully-qualified module name.
        module_name,

        # Module loader. Since this function dynamically creates this module,
        # this module is *NOT* loadable.
        #
        # Note that this parameter should technically only be "None" for
        # namespace packages. Unfortunately, there appears to exist no standard
        # loader class that effectively reduces to a noop. Ergo, "None" it is.
        None,
    )

    # Module defined by this specification to be returned.
    #
    # Note that modules may also be created by directly instantiating the
    # low-level "types.ModuleType" class but that doing so is officially
    # discouraged as "...spec is used to set as many import-controlled
    # attributes on the module as possible."
    module = importlib_util.module_from_spec(module_spec)

    # Set the name of the parent module of this module if any *OR* "None"
    # otherwise. Note that this special attribute could technically also be set
    # by passing the positional "parent" argument to the ModuleSpec.__init__()
    # method called above, but that doing so is complicated by:
    #
    # * That method's refusal to accept a keyword "parent" argument.
    # * The "ModuleSpec" class' refusal to allow callers to externally set the
    #   "parent" instance variable.
    module.__package__ = parent_module_name

    # If this module is to be prepopulated with attributes, do so.
    if module_attr_name_to_value is not None:
        module.__dict__.update(module_attr_name_to_value)

    # If this module is to be importable, register this module with the
    # standard "sys.modules" dictionary.
    if is_importable:
        sys.modules[module_name] = module

    # Return this module.
    return module



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Fixtures and fixture classes efficiently exercising a single subcommand of the
BETSE CLI in the active Python interpreter.
'''

#FIXME: We've added a new "--dist-dir" option to our "freeze_*" family of
#setuptools subcommands. Now, let's add support for fixtures running these
#subcommands. Is there any means of programmatically running setuptools
#subcommands in the current Python process? We suspect not, sadly.

#FIXME: Add seemless support for exercising all CLI-specific functional tests
#leveraging the "betse_cli" fixture against a frozen rather than unfrozen
#version of BETSE's CLI. Do *NOT* parametrize this fixture to unconditionally
#run each such test against both frozen and unfrozen versions of BETSE's CLI, as
#doing so would substantially increase space and time complexity with little to
#no tangible gain. Instead:
#
#* Define a new "--is-frozen" command-line option specific to our setuptools
#  "test" subcommand, defaulting to disabled.
#* When this option is *NOT* passed:
#  * One and only one PyInstaller test should be run -- say, test_pyi_try().
#    This test is intended only to provide a coarse-grained sanity check of
#    whether or not:
#    * BETSE is actually freezable under PyInstaller.
#    * The resulting executable is successfully simulatable with the default
#      simulation configuration via "betse try".
#  * Consequently, this test should (in order):
#    1. Freeze BETSE in one-dir mode (which requires no additional compression
#       and decompression steps and hence is slightly more time efficient for
#       our simple purposes) into a temporary subdirectory.
#    2. Run "betse try" against this frozen executable.
#  * All other CLI tests should be run as is against the unfrozen version of
#    BETSE importable by the active Python interpreter.
#* When this option is passed:
#  * The aforementioned test_pyi_try() test should *NOT* be run, as doing so
#    would be wholly redundant.
#  * BETSE should be frozen in one-dir mode, as discussed above.
#  * All other CLI tests should be run against this frozen executable rather
#    than the unfrozen, importable version of BETSE.
#
#Hence, the "--is-frozen" CLI option serves as a high-level switch fundamentally
#changing testing behaviour. The intention, of course, is that this option would
#only be passed immediately before producing an official new frozen version of
#BETSE. This provides a sanity check on frozen executable behaviour otherwise
#difficult (if not infeasible) to do manually.
#FIXME: This is great. We'd only like to make one improvement to the
#aforementioned PyInstaller testing support: marks. Use them. Explicit is
#typically better than implicit. In particular:
#
#* Define a new "betse_test.mark.able" submodule.
#* Define a new "@freezable" mark decorator in this submodule.
#* Decorate all tests suitable for running under a frozen copy of BETSE with
#  this decorator. This should probably *ONLY* include:
#  * All functional CLI tests, excluding "test_pyi_try". Everything else (e.g.,
#    unit tests) are *NOT* suitable for running frozen.
#FIXME: Urgh. That's great, but it's still not quite enough. Why? Simulation
#configurations. We currently create them by importing BETSE packages, which
#clearly won't work for frozen commands. Instead, if running frozen, we'll need
#to just call "betse config". Simple, if tedious.

# ....................{ IMPORTS                            }....................
from betse.util.os.command import cmdexit
from betse.util.type.types import type_check
from pytest import fixture

# ....................{ CONSTANTS                          }....................
_CLI_OPTIONS_MANDATORY = (
    '--verbose',
    '--log-level=none',
    '--matplotlib-backend=agg',
)
'''
Tuple of all failure-friendly command-line options unconditionally passed to
all invocations of the BETSE CLI by functional tests.

To improve debuggability for failing tests, these options are unconditionally
passed by :class:`CLITester` instances created by the :func:`betse_cli`
fixture:

* ``--verbose``, logging low-level debugging messages to stdout, which
  ``py.test`` captures for all tests and displays for all failing tests.
* ``--log-level=none``, redirecting all log messages to either stdout or stderr
  but *not* a logfile. While ``py.test`` can be configured to capture logfile
  messages, doing so sanely is complicated by the fact that ``py.test`` already
  captures both stdout and stderr by default. There is no advantage to
  recapturing logfile messages already logged to either stdout or stderr. So,
  tests avoid doing so entirely.
* ``--matplotlib-backend=agg``, enabling the default non-interactive matplotlib
  backend *guaranteed* to be usable on all platforms. By default, matplotlib
  enables an interactive backend (e.g., ``Qt5Agg``) inhibiting sane test
  automation.
'''

# ....................{ CLASSES                            }....................
class CLITester(object):
    '''
    BETSE CLI test runner, efficiently testing a single subcommand of the
    official BETSE CLI (i.e., ``betse``) in the active Python interpreter.

    Simple functional fixtures (e.g., ``betse_cli``) often return instances of
    this class to other fixtures and tests exercising a single facet of the
    BETSE CLI.

    Command Execution
    ----------
    For both efficiency and reliably, this runner does *not* actually fork this
    subcommand as a separate process. Doing so would introduce installation
    complications, portability concerns, and non-reproduceable edge-cases
    (e.g., conflicting versions of the same ``betse`` command in the current
    ``${PATH}``). Instead, this runner:

    * Imports the :mod:`betse.__main__` module implementing the BETSE CLI.
    * Passes this module's :func:`betse.__main__.main` function the passed
      arguments extended by the mandatory arguments defined by the
      :data:`_CLI_OPTIONS_MANDATORY` tuple global.
    '''

    # ..................{ RUNNERS                            }..................
    @type_check
    def run(self, *args: str) -> None:
        '''
        Run the BETSE CLI with the passed positional string arguments, extended
        by the mandatory positional string arguments defined by the
        :data:`_CLI_OPTIONS_MANDATORY` tuple global.

        Parameters
        ----------
        args : Tuple[str]
            Tuple of zero or more arguments to be passed to this entry point,
            corresponding exactly to the set of command-line arguments accepted
            by the external command for the BETSE CLI (i.e., ``betse``).

        See Also
        ----------
        ``betse --help``
            Further details on arguments accepted by the BETSE CLI.
        '''

        # Defer heavyweight imports to their point of use.
        from betse.__main__ import main
        from betse.util.app.meta import appmetaone

        # Deinitialize this application and hence the previously initialized
        # application metadata singleton if any. If this is *NOT* done, the
        # call to the main() function below will raise an exception on
        # attempting to reinstantiate this singleton.
        appmetaone.deinit()

        # Prefixed this argument list by failure-friendly options.
        args_evolved = _CLI_OPTIONS_MANDATORY + args
        # print('BETSE arg list: {}'.format(arg_list))

        # Run the BETSE CLI subcommand corresponding to these arguments,
        # capturing the exit status of that subcommand for testing.
        exit_status = main(args_evolved)

        # If this exit status signifies failure, fail the current test.
        assert cmdexit.is_success(exit_status), (
            'BETSE CLI failed with exit status {} '
            'given arguments: {}'.format(exit_status, args_evolved))

# ....................{ FIXTURES                           }....................
# Test-scope fixture creating and returning a new object for each unique test.
@fixture
def betse_cli() -> CLITester:
    '''
    Fixture returning a test-specific object suitable for running arbitrary
    BETSE CLI subcommands required by the current fixture or test.

    Returns
    ----------
    CLITester
        Object running arbitrary BETSE CLI subcommands.
    '''

    # Print a newline, ensuring that the first line of stderr or stdout printed
    # by the BETSE CLI is visually demarcated from prior py.test output.
    print()

    # Create and return a new BETSE CLI test runner.
    return CLITester()
